; *****************************************************************************
; *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
; *                                                                            
; * x-symbol-holocl.el  - HOL-OCL x-symbol minor mode
; * Copyright (C) 2005 Achim D. Brucker <brucker@inf.ethz.ch>
; *
; * This file is part of HOL-OCL.
; *
; * HOL-OCL is free software; you can redistribute it and/or modify it under   
; * the terms of the GNU General Public License as published by the Free       
; * Software Foundation; either version 2 of the License, or (at your option)  
; * any later version.                                                         
; *                                                                            
; * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
; * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
; * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
; * details.                                                              
; * 
; * You should have received a copy of the GNU General Public License along    
; * with this program; if not, write to the Free Software Foundation, Inc.,    
; * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
; ******************************************************************************


(provide 'x-symbol-holocl)

(require 'x-symbol-vars)
(require 'x-symbol-isar)


(defcustom x-symbol-holocl-header-groups-alist
  (append '(
;  ("Math-OCL (OLD)" MathOclSymbol MathOclRelation MathOclArrow)
  ("Math-OCL (OclAny)" MathOclOclAny)
  ("Math-OCL (Collection)" MathOclCollection)
  ("Math-OCL (Logic)" MathOclLogic)
  ("Math-OCL (Numerals)" MathOclNumerals)
  ("Math-OCL (String)" MathOclString)
  ("Math-OCL (OclVoid)" MathOclVoid)
  ("HOL-OCL (Lifting)" HolOclLifting)
  )
   x-symbol-header-groups-alist
  )
   "*If non-nil, used in isasym specific grid/menu.
See `x-symbol-header-groups-alist'."
  :group 'x-symbol-holocl
  :group 'x-symbol-input-init
  :type 'x-symbol-headers)


(defvar x-symbol-holocl-required-fonts '(x-symbol-holocl-font))


(defvar x-symbol-holocl-modeline-name "HOL-OCL")



(defcustom x-symbol-holocl-electric-ignore
  "[:'][A-Za-z]\\|<=\\|\\[\\[\\|\\]\\]\\|~="
  "*Additional holocl version of `x-symbol-electric-ignore'."
  :group 'x-symbol-holocl
  :group 'x-symbol-input-control
  :type 'x-symbol-function-or-regexp)


(defvar x-symbol-holocl-required-fonts nil
  "List of features providing fonts for language `holocl'.")

(defvar x-symbol-holocl-extra-menu-items nil
  "Extra menu entries for language `holocl'.")


(defvar x-symbol-holocl-token-grammar
  '(x-symbol-make-grammar
    :encode-spec (((t . "\\\\")))
    :decode-regexp "\\\\+<[A-Za-z0-9]+>"
    :input-spec nil
    :token-list x-symbol-holocl-token-list))

(defun x-symbol-holocl-token-list (tokens)
  (mapcar (lambda (x) (cons x t)) tokens))
 
(defvar x-symbol-holocl-user-table nil
 "User table defining holocl commands, used in `x-symbol-holocl-table'.")

(defvar x-symbol-holocl-generated-data nil "Internal.")


;;;===========================================================================
;;;  Charsym Info
;;;===========================================================================


(defface x-symbol-holocl-symbol-face
  '((((class color) (background light))
     (:foreground "purple3")))
  "*Face, normally used for tokens only allowed in TeX's math mode.
Used in `x-symbol-tex-class-face-alist'."
  :group 'x-symbol-tex
  :group 'x-symbol-info-general)



(defcustom x-symbol-holocl-class-alist
  '((MathOcl)
    (MathOclSymbol)
    (MathOclOclAny)
    (MathOclLogic)
    (MathOclCollection)
    (MathOclRelation "no holocl Symbol" (blue x-symbol-emph-info-face))
    (VALID "holocl Symbol" (x-symbol-info-face))
    (INVALID "no holocl Symbol" (red x-symbol-info-face))
)
"Alist for holocl's token classes displayed by info in echo area.
See `x-symbol-language-access-alist' for details."
  :group 'x-symbol-texi
  :group 'x-symbol-info-strings
;;  :set 'x-symbol-set-cache-variable   [causes compile error]
  :type 'x-symbol-class-info)


(defcustom x-symbol-holocl-class-face-alist
  '((MathOclSymbol x-symbol-holocl-symbol-face (x-symbol-holocl-symbol-face))
    (MathOclRelation  x-symbol-holocl-symbol-face (x-symbol-info-face)))

  "Alist for holocl's color scheme in TeXinfo's grid and info.
See `x-symbol-language-access-alist' for details."
  :group 'x-symbol-holocl
  :group 'x-symbol-input-init
  :group 'x-symbol-info-general
;;  :set 'x-symbol-set-cache-variable   [causes compile error]
  :type 'x-symbol-class-faces)



;;;=================================================================
;;;  The tables
;;;=================================================================

(defvar x-symbol-holocl-case-insensitive nil)
(defvar x-symbol-holocl-token-shape nil)
(defvar x-symbol-holocl-input-token-ignore nil)
(defvar x-symbol-holocl-token-list 'identity)

;(defvar OCLexists "->exists")
(defvar x-symbol-holocl-symbol-table      ; symbols (holocl font)
  '(
    (OclImplies     nil "\\<MathOclImplies>")
    (OclVImplies nil "\\<MathOclVImplies>")
    (OclVVImplies nil "\\<MathOclVVImplies>")
    (OclSImplies    nil "\\<MathOclSImplies>")
;    (Ocl_          nil "\\<OCLif>")
    (OclLiftLeft    nil "\\<HolOclLiftLeft>")
    (OclLiftRight   nil "\\<HolOclLiftRight>")

    (OclDropLeft    nil "\\<HolOclDropLeft>")
    (OclDropRight   nil "\\<HolOclDropRight>")
    (OclFloorLeft   nil "\\<MathOclFloorLeft>")
    (OclFloorRight  nil "\\<MathOclFloorRight>")
    (OclRoundLeft   nil "\\<MathOclRoundLeft>")
    (OclRoundRight  nil "\\<MathOclRoundRight>")
;    (OCL_ nil "\\<OCLif>")
;    (OCL_ nil "\\<OCLif>")
    (OclIf          nil "\\<MathOclIf>")
    (OlcThen        nil "\\<MathOclThen>")
    (OclElse        nil "\\<MathOclElse>")
    (OclEndif       nil "\\<MathOclEndif>")
    (OclLet         nil "\\<MathOclLet>")
    (OclIn          nil "\\<MathOclIn>")
    (OclEnd         nil "\\<MathOclEnd>")
    (OclFalse       nil "\\<MathOclFalse>")
    (OclTrue        nil "\\<MathOclTrue>")
    (OclUndefined   nil "\\<MathOclUndefined>")
    (OclStrictEq    nil "\\<MathOclStrictEq>")
    (OclStrictNeq   nil "\\<MathOclStrictNeq>")
    (OclStrictSVE   nil "\\<MathOclStrictValueEq>")
    (OclStrongSVE   nil "\\<MathOclStrongValueEq>")
    (OclStrictDVE   nil "\\<MathOclStrictDeepValueEq>")
    (OclStrongDVE   nil "\\<MathOclStrongDeepValueEq>")

    (OclStrongeq     nil "\\<MathOclStrongEq>")
    (OclStrongNeq    nil "\\<MathOclStrongNEq>")
    (OclRefEq        nil "\\<MathOclRefEq>")
    (OclRefNeq       nil "\\<MathOclRefNeq>")
    (OclIsDefined    nil "\\<MathOclIsDefined>")
    (OclIsUndefined  nil "\\<MathOclIsUndefined>")
    (OclAsType       nil "\\<MathOclAsType>")
    (OclIsType       nil "\\<MathOclIsType>")
    (OclIsKindOf     nil "\\<MathOclIsKindOf>")
    (OclAllInstances nil "\\<MathOclAllInstances>")

    (OclPlus         nil "\\<MathOclPlus>")
    (OclMinus        nil "\\<MathOclMinus>")
    (OclMult         nil "\\<MathOclMult>")
    (OclNegate       nil "\\<MathOclNegate>")
    (OclDivision     nil "\\<MathOclDivision>")
    (OclAbsLeft      nil "\\<MathOclAbsLeft>")
    (OclAbsRight     nil "\\<MathOclAbsRight>")
    (OclSizeRight    nil "\\<MathOclSizeRight>")
    (OclSizeLeft     nil "\\<MathOclSizeLeft>")
    (OclMax          nil "\\<MathOclMax>")

    (OclMin          nil "\\<MathOclMin>")
    (OclLess         nil "\\<MathOclLess>")
    (OclGreater      nil "\\<MathOclGreater>")
    (OclLessEq       nil "\\<MathOclLe>")
    (OclGreaterEq    nil "\\<MathOclGe>")
;    (Ocl          nil "\\<MathOcl>")
    (OclDiv          nil "\\<MathOclDiv>")
    (OclMod          nil "\\<MathOclMod>")
    (OclConcat       nil "\\<MathOclConcat>")
    (OclSubstring    nil "\\<MathOclSubstring>")

    (OclToInteger    nil "\\<MathOclToInteger>")
    (OclToReal       nil "\\<MathOclToReal>")
    (OclToUpper      nil "\\<MathOclToUpper>")
    (OclToLower      nil "\\<MathOclToLower>")

    (OclValid        nil "\\<MathOclValid>")
    (OclOr           nil "\\<MathOclOr>")
    (OCLSor          nil "\\<MathOclSor>")
    (OCLAnd          nil "\\<MathOclAnd>")
    (OCLSAnd         nil "\\<MathOclSand>")
    (OCLXor          nil "\\<MathOclXor>")

    (OCLSxor        nil "\\<MathOclSxor>")
    (OclNot         nil "\\<MathOclNot>")
    (OclIncludes    nil "\\<MathOclIncludes>")
    (OCLExcludes    nil "\\<MathOclExcludes>")
    (OclCount       nil "\\<MathOclCount>")
    (OclIndludesAll nil "\\<MathOclIncludesAll>")
 ;   (OCL_ nil "\\<MathOclSupSetEq>")
    (OclExcludesAll nil "\\<MathOclExcludesAll>")
    (OclIsEmpty     nil "\\<MathOclIsEmpty>")
    (OclIsNotEmpty  nil "\\<MathOclIsNotEmpty>")

    (OclSum           nil "\\<MathOclSum>")
    (OclProduct       nil "\\<MathOclProduct>")
    (OclEmptySet      nil "\\<MathOclMtSet>")
    (OclUnion         nil "\\<MathOclUnion>")
    (OclIntersection  nil "\\<MathOclIntersection>")
    (OclComplement    nil "\\<MathOclComplement>")
    (OclSymDifference nil "\\<MathOclSymDifference>")
    (OclFlattenLeft   nil "\\<MathOclFlattenLeft>")
    (OclFlattenRight  nil "\\<MathOclFlattenRight>")
    (OclAsSet         nil "\\<MathOclAsSet>")

    (OclAsOrderedSet  nil "\\<MathOclAsOrderedSet>")
    (OclAsBag         nil "\\<MathOclAsBag>")
    (OclAsSequence    nil "\\<MathOclAsSequence>")
 ;   (Ocl_ nil "\\<MathOcl>")
    (OclAppend        nil "\\<MathOclAppend>")
    (OclPrepend       nil "\\<MathOclPrepend>")
    (OclInsertAt      nil "\\<MathOclInsertAt>")
;    (Ocl nil "\\<MathOcl>")
    (OclSubOrderedSet nil "\\<MathOclSubOrderedSet>")
    (OclAt            nil "\\<MathOclAt>")

    (OclIndexOf       nil "\\<MathOclIndexOf>")
    (OclFirst         nil "\\<MathOclFirst>")
    (OclLast          nil "\\<MathOclLast>")
    (OclSubSequence   nil "\\<MathOclSubSequence>")
    (OclBagLeft       nil "\\<MathOclBagLeft>")
    (OclBagRight      nil "\\<MathOclBagRight>")
    (OclEmptyBag      nil "\\<MathOclMtBag>")
    (OclSequenceLeft  nil "\\<MathOclSequenceLeft>")
    (OclSequenceRight nil "\\<MathOclSequenceRight>")
    (OclEmptySequence nil "\\<MathOclMtSequence>")

    (OclSetLeft       nil "\\<MathOclSetLeft>")
    (OclSetRight      nil "\\<MathOclSetRight>")
;    (Ocl_ nil "\\<MathOcl>")
    (OclOrderedSetLeft  nil "\\<MathOclOrderedSetLeft>")

    (OclOrderedSetRight nil "\\<MathOclOrderedSetRight>")
    (OclEmptyOrderedSet nil "\\<MathOclMtOrderedSet>")
    (OclForall          nil "\\<MathOclForAll>")
    (OclExists          nil "\\<MathOclExists>")
    (OclIsUnique        nil "\\<MathOclIsUnique>")
    (OclOne             nil "\\<MathOclOne>")

    (OclAny             nil "\\<MathOclAny>")
    (OclCollect         nil "\\<MathOclCollect>")
    (OclSelect          nil "\\<MathOclSelect>")
    (OclReject          nil "\\<MathOclReject>")
    (OclSelectLeft      nil "\\<MathOclSelectLeft>")
    (OclSelectRight     nil "\\<MathOclSelectRight>")
    (OclRejectRight     nil "\\<MathOclRejectRight>")
    (OclRejectLeft      nil "\\<MathOclRejectLeft>")
    (OclCollectLeft     nil "\\<MathOclCollectLeft>")
    (OclCollectRight    nil "\\<MathOclCollectRight>")

    (OclCollectNestedLeft  nil "\\<MathOclCollectNestedLeft>")
    (OclCollectNestedRight nil "\\<MathOclCollectNestedRight>")
    (OclSortedBy           nil "\\<MathOclSortedBy>")
   ))

 
(defvar x-symbol-holocl-table
   (append
    x-symbol-holocl-user-table
    x-symbol-holocl-symbol-table
    x-symbol-isabelle-table
   ))

;(defvar x-symbol-header-groups-alist
;   (append
;    x-symbol-holocl-header-groups-alist
;    x-symbol-header-groups-alist
;   ))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; User-level settings for X-Symbol
;;
;; this is MODE-ON CODING 8BITS UNIQUE SUBSCRIPTS IMAGE
(defcustom x-symbol-holocl-auto-style
  '((proof-ass x-symbol-enable)	 ; MODE-ON: whether to turn on interactively
    nil   ;; x-symbol-coding
    'null ;; x-symbol-8bits	   [NEVER want it; null disables search]
    nil   ;; x-symbol-unique
    t     ;; x-symbol-subscripts
    nil)  ;; x-symbol-image
  "Variable used to document a language access.
See documentation of `x-symbol-auto-style'."
  :group 'x-symbol-holocl
  :group 'x-symbol-mode
  :type 'x-symbol-auto-style)



;(setq
; proof-xsym-activate-command
; (isar-markup-ml
;  "print_mode := ([\"OCL\",\"xsymbols\", \"symbols\"] @ ! print_mode)")
; proof-xsym-deactivate-command
; (isar-markup-ml
;  "print_mode := (Library.gen_rems (op =) (! print_mode, [\"OCL\",\"xsymbols\", \"symbols\"]))"))

(setq x-symbol-unique t)




(setq font-lock-auto-fontify t)
(setq font-lock-maximum-decoration t)

(setq x-symbol-token-language 'holocl)
(defpgdefault x-symbol-language 'holocl)

   (x-symbol-decode)
