(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * encoding_example.thy - 
 * Copyright (C) 2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This File is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


header {* Encoding Recursive Data Structures *}

theory tree_aux
imports 
  digraph
begin

lemma Node_inv_0 :  
  "(self \<in> (Node.model.set \<tau>)) = (Node.model.hol_invariant \<tau> (Node.model.set \<tau>) self)"
  apply(subst Node.model.set)
  apply(subst gfp_unfold[OF level1.tree.Node.model.set_fun.mono],simp)
  apply(subst Node.model.set_fun)
  apply(simp add:  Node.model.set)
  done

lemma rapperswil: "
  Node.model.hol_invariant =
  (\<lambda>\<tau> C obj.
    model.hol_invariant.init_i \<tau> C obj \<and>
    model.hol_invariant.init_left \<tau> C obj \<and>
    model.hol_invariant.init_right \<tau> C obj \<and>
    model.hol_invariant.inv_3 \<tau> C obj \<and>
   (
   ((level1.tree.Node.left obj \<tau>) \<in> ((\<lambda> f. f \<tau> ) ` C)) 
  \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.tree.Node.left obj)
   )\<and> (

   ((level1.tree.Node.right obj \<tau>) \<in> ((\<lambda> f. f \<tau> ) ` C))
  \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.tree.Node.right obj)
  ) \<and>
   Node.model.hol_invariant.check_parent_inv \<tau> C obj
)
"
  apply(unfold Node.model.hol_invariant)
  apply(rule ext)+
  apply(auto)
  apply(unfold model.hol_invariant.typeconstraint_left_def)
  apply(unfold level1.tree.Node.left_def)
  apply(unfold model.hol_invariant.typeconstraint_right_def)
  apply(unfold level1.tree.Node.right_def)
  apply(simp_all)
  apply(erule disjE)+
  apply(simp_all)
  apply(auto)
  done


lemma image_cong:
"\<lbrakk>!!x . x \<in> S \<Longrightarrow> f x = g x\<rbrakk> \<Longrightarrow> f ` S = g ` S"
by (auto simp: image_def)

lemma Cnode_inv_0 :  
      "(self \<in> (Cnode.model.set \<tau>)) = (Cnode.model.hol_invariant \<tau> (Cnode.model.set \<tau>) self)"
apply(subst Cnode.model.set)
apply(subst gfp_unfold[OF level1.tree.Cnode.model.set_fun.mono],simp)
apply(subst Cnode.model.set_fun)

apply(simp add:  Cnode.model.set Node.model.set_fun)
done
lemma horgen: "
Node.model.hol_invariant \<tau>  (Node.model.set \<tau> ) obj =
(
   model.hol_invariant.init_i \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_left \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.init_right \<tau> (Node.model.set \<tau>) obj \<and>
   model.hol_invariant.inv_3 \<tau> (Node.model.set \<tau>) obj \<and>
   (
   (\<tau>  \<MathOclValid> (level1.tree.Node.left obj) \<MathOclIncludes> Node.model.characteristic_set)  
  \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.tree.Node.left obj)
   )\<and> (

   (\<tau>  \<MathOclValid> (level1.tree.Node.right obj) \<MathOclIncludes> Node.model.characteristic_set) 
  \<or> \<tau> \<MathOclValid> \<MathOclIsUndefined> (level1.tree.Node.right obj)
  ) \<and>
   Node.model.hol_invariant.check_parent_inv \<tau> (Node.model.set \<tau>) obj
)
"
  apply(simp add: rapperswil Node.model.characteristic_set_def)
  apply(auto)
  apply(rule OCL_Set.includes_charn2, simp_all add: DEF_def localValidDefined2sem)+
  apply(drule OCL_Set.includes_charn1, simp)+
  done

lemmas foo [intro!] = level1.tree.Node.cp_left [THEN cp_compose2]


lemma def_coinductZ:
     "[| A==gfp(f);  mono(f);  X \<subseteq> f(X Un A) |] ==> X \<subseteq> A"
apply(auto intro!: coinduct)
done


end
