(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Any.thy - Type Definition for the Referential OclAny
 * Copyright (C) 2003-2006           Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* Type Definition for OclAny *}

theory OCL_OclAny_type
imports 
  "$HOLOCL_HOME/src/Lifting" 
begin
text_raw{* \label{thy:OCL_OclAny_type} *}
typedecl 
      oid 

arities 
      oid :: bot

datatype OclAny_key = OclAny   

types  OclAny_type = "OclAny_key \<times> oid"
text {* 
  This definition of @{typ "OclAny_type"} results
  in a  ``referential universe'', i.e., a universe where a referential 
  equality is present.  A referential equality is the strong equality 
  for ``boxed types'' (that is Real, Boolean, String) and an equality on 
  the reference to a value. Since our universe construction is designed to 
  mirror the subtype discipline of \OCL, i.e., objects are records of a 
  certain type, the reference to an object must be stored in the object 
  itself in our model, namely at the first position of the record.

  If OCL expressions refer to states, where all objects contain the
  reference to itself in the state, the logical equality co-incides
  with a referential equality.

*}

types 
      'a OclAny_ext = "(OclAny_type \<times> 'a up)"
      'a OclAny_0   = "( 'a OclAny_ext) up"
    
text {*
  Alternative we could define which would allow an additional
  `exception' element with a non-strict behavior. This would
  model something like a null reference (or pointer) known
  from object-oriented programming languages.
*}

translations 
      "a OclAny_0"  <= (type) "((OclAny_key * oid ) * a up)up"

types     ('st,'a)OclAny   = "('st, 'a OclAny_0) VAL"

lemma [simp]: "DEF (obj\<Colon>(('a)OclAny_0)) \<Longrightarrow> 
             (fst(fst(drop(obj))) = OclAny_key.OclAny)"
  apply (auto simp:  DEF_def not_down_exists_lift) 
  apply (case_tac a)
  apply (assumption)
  done

end
