(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Integer.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Integer *}  
theory OCL_Integer 
imports 
  "$HOLOCL_HOME/src/library/basic/OCL_Numerals" 
begin
text_raw{* \label{thy:OCL_Integer} *}


constdefs
  is_Integer_0    :: "'a U \<Rightarrow> bool"
  "is_Integer_0   \<equiv> sum_case (\<lambda>x. False) 
                             (sum_case (\<lambda>x. False)
                             (sum_case (\<lambda>x. True)(\<lambda>x. False)))"
  get_Integer_0   :: "'a U \<Rightarrow> Integer_0"
  "get_Integer_0  \<equiv> sum_case (\<lambda>x. \<some> x. True)
                              (sum_case  (\<lambda>x. \<some>  x. True)
                              (sum_case (\<lambda>x. x)(\<lambda>x. \<some> x. True)))"
  mk_Integer_0    :: "Integer_0 \<Rightarrow> 'a U"
  "mk_Integer_0   \<equiv> Inr \<circ> Inr \<circ> Inl"

lemma get_mk_Integer_id_0: "get_Integer_0(mk_Integer_0 x) = x"
  apply (simp add: get_Integer_0_def mk_Integer_0_def)
  done
lemma is_mk_Integer_0: "is_Integer_0(mk_Integer_0 x) = True"
  apply (simp add: is_Integer_0_def mk_Integer_0_def)
  done
lemma mk_get_Integer_id_0: "is_Integer_0 x \<Longrightarrow> mk_Integer_0(get_Integer_0 x) = x"
  apply (simp add: get_Integer_0_def mk_Integer_0_def is_Integer_0_def )
  apply (case_tac x, simp, simp)
  apply (case_tac b, simp,simp)
  apply (case_tac ba, simp,simp)
  done


constdefs
   Integer_0 :: "('st, Integer_0 Set_0) VAL"
  "Integer_0 \<equiv> lift0(Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>)"

defs (*of overloaded constants *)
  Zero_ocl_int_def: "0 \<equiv> lift\<^sub>0( \<HolOclLiftLeft>0\<Colon>int\<HolOclLiftRight> )"
  One_ocl_int_def:  "1 \<equiv> lift\<^sub>0( \<HolOclLiftLeft>1\<Colon>int\<HolOclLiftRight> )"


lemma OCL_is_def_Zero_ocl_int [simp]:
  "\<MathOclIsDefined> (0::'a Integer) = \<MathOclTrue>"
  apply(rule ext)
  apply(simp add: OclIsDefined_def Zero_ocl_int_def OclTrue_def ss_lifting')
  done

lemma OCL_is_def_One_ocl_int [simp]:
  "\<MathOclIsDefined> (1::'a Integer) = \<MathOclTrue>"
  apply(rule ext)
  apply(simp add: OclIsDefined_def One_ocl_int_def OclTrue_def ss_lifting')
  done


defs (overloaded) 
  number_of_def : "(number_of:: bin \<Rightarrow> 'a Integer)  
                        \<equiv>  \<lambda> b. lift\<^sub>0( \<HolOclLiftLeft>(number_of::bin \<Rightarrow> int) b\<HolOclLiftRight> )"


defs 
  plus_def:   "op + \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. strictify(\<lambda> y. 
                          \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> )+( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"

  minus_def:  "op - \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. strictify(\<lambda> y. 
                          \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> )-( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"

  times_def:  "op * \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. strictify(\<lambda> y. 
                          \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> )*( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"

ocl_setup_op [plus, minus, times]


lemma plus_zero: "(X::'a Integer) + 0 = X"
  by(rule ext, simp add: plus_def Zero_ocl_int_def ss_lifting')

lemma plus_commute: "(X::'a Integer) + Y = Y + X"
  by(rule ext, simp add: plus_def ss_lifting')



defs
  OclLe_def:     "OclLe     \<equiv> lift\<^sub>2(strictify(\<lambda>x\<Colon>Integer_0. 
                              strictify(\<lambda>y.  \<HolOclLiftLeft> ( \<HolOclDropLeft>x\<HolOclDropRight> ) <= ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  OclLess_def:   "OclLess   \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. 
                              strictify(\<lambda> y. \<HolOclLiftLeft> ( \<HolOclDropLeft>x\<HolOclDropRight> ) <  ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  OclGe_def:     "OclGe     \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. 
                              strictify(\<lambda> y. \<HolOclLiftLeft> \<not>( \<HolOclDropLeft>x\<HolOclDropRight> ) < ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"
  OclGreater_def:"OclGreater\<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. 
                              strictify(\<lambda> y. \<HolOclLiftLeft> \<not>( \<HolOclDropLeft>x\<HolOclDropRight> )<=( \<HolOclDropLeft>y\<HolOclDropRight> ) \<HolOclLiftRight> )))"

ocl_setup_op [OclLe, OclLess, OclGe, OclGreater]


lemma OclGe_elim[simp] : "OclGe (X::'a Integer) Y =  OclLe Y X"
apply(rule ext)
apply(case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(X)",simp_all, ocl_hypsubst [2], simp_all)
apply(case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(Y)",simp_all, ocl_hypsubst [2], simp_all)
apply(simp add:  OclGe_def OclLe_def
                 OclIsDefined_def strictify_def OclLocalValid_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclUndefined_def DEF_def) 
apply(simp add: Orderings.linorder_not_less)
done


lemma OclGreater_elim[simp] : "OclGreater (X::'a Integer) Y =  OclLess Y X"
apply(rule ext)
apply(case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(X)",simp_all, ocl_hypsubst [2], simp_all)
apply(case_tac "x  \<MathOclValid>  \<MathOclIsDefined>(Y)",simp_all, ocl_hypsubst [2], simp_all)
apply(simp add:  OclGreater_def OclLess_def
                 OclIsDefined_def strictify_def OclLocalValid_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclUndefined_def DEF_def) 
apply(simp add: Orderings.linorder_not_le)
done


lemma OclLe_compute [simp]: 
      "OclLe((number_of:: bin \<Rightarrow> 'a Integer) a)((number_of:: bin \<Rightarrow> 'a Integer) b) = 
       (if \<not> neg ((number_of:: bin \<Rightarrow> int) (bin_add b (bin_minus a)))then \<MathOclTrue> else \<MathOclFalse>) "
apply (rule ext)
apply (simp add: number_of_def OclLe_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done


lemma OclLess_compute [simp]: 
      "OclLess((number_of:: bin \<Rightarrow> 'a Integer) a)((number_of:: bin \<Rightarrow> 'a Integer) b) = 
       (if neg ((number_of:: bin \<Rightarrow> int) (bin_add a (bin_minus b)))then \<MathOclTrue> else \<MathOclFalse>) "
apply (rule ext)
apply (simp add: number_of_def OclLess_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done


consts

  OclMod    :: "['a Integer, 'a Integer] \<Rightarrow> 'a Integer"    
               ("_ ->mod'( _ ')" [66,65]65) 

  OclDiv    :: "['a Integer, 'a Integer] \<Rightarrow> 'a Integer"    
               ("_ ->div'( _ ')" [66,65]65) 

  OclRound  :: "'a Integer \<Rightarrow> 'a Integer"             
               ("_ ->round'( ')" [66])		

  OclFloor  :: "'a Integer \<Rightarrow> 'a Integer"             
               ("_ ->floor'( ')" [66])		
 
  OclMin    :: "['a Integer, 'a Integer] \<Rightarrow> 'a Integer"    
               ("_ ->min'( _ ')" [66,65]65) 

  OclMax    :: "['a Integer, 'a Integer] \<Rightarrow> 'a Integer"    
               ("_ ->max'( _ ')" [66,65]65) 

  OclAbs    :: "'a Integer \<Rightarrow> 'a Integer"             
               ("_ ->abs'( ')" [66])		

  OclNegative :: "'a Integer \<Rightarrow> 'a Integer"             
                 ("- _" [66])	
  
 

defs
  OclRound_def: "SELF ->round() \<equiv> SELF"  (* HACK *)
  OclFloor_def: "SELF ->floor() \<equiv> SELF"  (* HACK *)


defs
  OclMod_def: "OclMod \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. strictify(\<lambda> y. 
                              if \<HolOclDropLeft>y\<HolOclDropRight> = 0 then down else
                              \<HolOclLiftLeft> ( \<HolOclDropLeft>x\<HolOclDropRight> ) mod ( \<HolOclDropLeft>y\<HolOclDropRight> ) \<HolOclLiftRight> )))"

  OclDiv_def: "OclDiv \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>Integer_0. strictify(\<lambda> y. 
                              if \<HolOclDropLeft>y\<HolOclDropRight> = 0 then down else
                              \<HolOclLiftLeft> ( \<HolOclDropLeft>x\<HolOclDropRight> ) div ( \<HolOclDropLeft>y\<HolOclDropRight> ) \<HolOclLiftRight> )))"

  OclMin_def: "SELF ->min(I) \<equiv> \<MathOclIf> (SELF \<MathOclLe> I) \<MathOclThen> SELF \<MathOclElse> I \<MathOclEndif>"
 
  OclMax_def: "SELF ->max(I) \<equiv> \<MathOclIf> (SELF \<MathOclGe> I) \<MathOclThen> SELF \<MathOclElse> I \<MathOclEndif>"
 
  OclAbs_def: "OclAbs \<equiv> lift\<^sub>1(strictify(\<lambda> x\<Colon>Integer_0. 
                            \<HolOclLiftLeft> abs ( \<HolOclDropLeft>x\<HolOclDropRight> ) \<HolOclLiftRight> ))"   

  OclNegative_def: "OclNegative \<equiv> lift\<^sub>1(strictify(\<lambda> x\<Colon>Integer_0. 
                            \<HolOclLiftLeft> ( - ( \<HolOclDropLeft>x\<HolOclDropRight> )) \<HolOclLiftRight> ))"   


ocl_setup_op  [OclMod, OclDiv, OclAbs, OclNegative]


section{* Computational Rules *}


text{* Foundational rule over Integer values in \OCL: *}

lemma Integer_values_defined[simp]: 
      "\<MathOclIsDefined>((number_of:: bin \<Rightarrow> 'a Integer) a) = \<MathOclTrue>"
apply (rule ext)
by    (simp add: number_of_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclUndefined_def DEF_def) 


text{* These two elementary cases are required by the
       standard: *}
lemma div_zero: "x ->div( 0 ) = \<MathOclUndefined>"
apply (rule ext)
apply (simp add: OclDiv_def Zero_ocl_int_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclUndefined_def DEF_def)
done

lemma mod_zero: "x ->mod( 0 ) = \<MathOclUndefined>"
apply (rule ext)
apply (simp add: OclMod_def Zero_ocl_int_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done


lemma plus_compute [simp]: 
      "((number_of:: bin \<Rightarrow> 'a Integer) a) + ((number_of:: bin \<Rightarrow> 'a Integer) b) = 
       number_of (bin_add a b)"
apply (rule ext)
apply (simp add: number_of_def plus_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done

lemma "3 + 4 = (7::'a Integer)" by simp
    
lemma times_compute [simp]: 
      "((number_of:: bin \<Rightarrow> 'a Integer) a) * ((number_of:: bin \<Rightarrow> 'a Integer) b) = 
       number_of (bin_mult a b)"
apply (rule ext)
apply (simp add: number_of_def times_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done
 



lemma minus_compute [simp]: 
      "((number_of:: bin \<Rightarrow> 'a Integer) a) - ((number_of:: bin \<Rightarrow> 'a Integer) b) = 
       (number_of ((bin_add a (bin_minus b))))"
apply (rule ext)
apply (simp add: number_of_def minus_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done


lemma strong_eq_compute [simp]: 
      " \<not> iszero((number_of::bin \<Rightarrow> int) (bin_add a (bin_minus b))) \<Longrightarrow>
        (((number_of::bin\<Rightarrow>'a Integer)a) \<MathOclStrongEq> ((number_of::bin \<Rightarrow>'a Integer)b)) 
        = \<MathOclFalse> "
apply (rule ext)
apply (simp add: number_of_def OclStrongEq_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done

lemma strict_eq_compute [simp]: 
      " \<not> iszero((number_of::bin \<Rightarrow> int) (bin_add a (bin_minus b))) \<Longrightarrow>
        (((number_of::bin\<Rightarrow>'a Integer)a) \<MathOclStrictEq> ((number_of::bin \<Rightarrow>'a Integer)b)) 
        = \<MathOclFalse> "
apply (rule ext)
apply (simp add: number_of_def OclStrictEq_def
                 OclIsDefined_def strictify_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def OclFalse_def OclUndefined_def DEF_def)
done

lemma "\<lbrakk> \<tau> \<MathOclValid> 3 \<MathOclStrongEq> C ; \<tau> \<MathOclValid> C + 2 \<MathOclStrictEq> (7::'a Integer)\<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> A \<MathOclAnd> B " 
apply ocl_hypsubst
apply simp
done

end 



