(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Collection.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)
header {* \OCL Collection *}

theory OCL_Collection 
imports 
  "$HOLOCL_HOME/src/library/basic/OCL_Integer"
begin
 text_raw{* \label{thy:OCL_Collection} *}
text{* 
  The purpose of this theory is to provide the syntactic framework for
  the abstract class \inlineocl+Collection+ in the sense of the
  standard.  It is defined by the Isabelle type class ``collection''.
  We require that collections must always have a bottom element.

  Unfortunately, this implementation by the type class ``collection'' is
  an approximation (a type constructor class would be more
  appropriate, but is not available in the Isabelle type
  system). Consequently, in lemmas the concrete type instances for
  operators resulting from the abstract class must be given; the types
  inferred by Isabelle automatically are too coarse and can lead to
  dead ends in proofs (definitions can not be unfolded). 
*}

section{*The Syntax of Abstract Operations*}

text{* 
  The core operation \emph{size} (which may be undefined if and only
  if the collection is infinitely large), \emph{select}, \emph{collect},
   \emph{includes} and \emph{any}.
*}

subsection {* Operations on all Collection Types *}
consts (* abstract collection operations *)
 
    OclSize        :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> '\<tau> Integer" 
    OclCount       :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> '\<tau> Integer" 
    OclIncludes    :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> '\<tau> Boolean" 
    OclExcludes    :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>) VAL] \<Rightarrow> '\<tau> Boolean" 
    OclIncluding   :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
    OclExcluding   :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
    OclFlatten     :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> ('\<tau>,'\<beta>::bot) VAL" 
    OclSum         :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> '\<tau> Integer"
    OclAsSet       :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> ('\<tau>,'\<beta>::collection) VAL" 
    OclAsSequence  :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> ('\<tau>,'\<beta>::collection) VAL" 
    OclAsBag       :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> ('\<tau>,'\<beta>::collection) VAL" 
    OclAsOrderedSet:: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> ('\<tau>,'\<beta>::collection) VAL" 
    OclIncludesAll :: "[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
    OclExcludesAll :: "[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
    OclIsEmpty     :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> '\<tau> Boolean"
    OclNotEmpty    :: " ('\<tau>,'\<alpha>::collection)VAL \<Rightarrow> '\<tau> Boolean" 
    OclComplement  :: " ('\<tau>,'\<beta>::collection)VAL \<Rightarrow> ('\<tau>,'\<beta>) VAL" 
    OclUnion       :: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] \<Rightarrow> ('\<tau>,'\<beta>)VAL"
    OclCollectionRange  :: "[('\<tau>,'\<alpha>) VAL,('\<tau>,'\<alpha>) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL" (* Syntax ??? *)

syntax  (* standard syntax *)
    "_OclSize_std"      :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Integer" 
		            ("_ \<OclSize>'(')" [66])

    "_OclCount_std"     :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] => '\<tau> Integer" 
		            ("_ \<OclCount>'(_')" [66,65]65)

    "_OclIncludes_std"  :: "[ ('\<tau>, '\<beta>::collection) VAL,('\<tau>, '\<alpha>) VAL] => '\<tau> Boolean" 
		            ("_ \<OclIncludes>'(_')" [66,65]65) 

    "_OclExcludes_std"  :: "[('\<tau>, '\<beta>::collection) VAL,('\<tau>, '\<alpha>::bot) VAL] => '\<tau> Boolean"
 		            ("_ \<OclExcludes>'(_')" [66,65]65)

    "_OclFlatten_std"   :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ \<OclFlatten>'(')" [66])

    "_OclSum_std"       :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Integer"
		            ("_ \<OclSum>'(')" [66])

    "_OclAsSet_std"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ \<OclAsSet>'(')" [66])

    "_OclAsSequence_std":: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ \<OclAsSequence>'(')" [66])

    "_OclAsBag_std"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ \<OclAsBag>'(')" [66])
 
    "_OclAsOrderedSet_std":: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ \<OclAsOrderedSet>'(')" [66])

    "_OclIncludesAll_std":: "[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>::collection)VAL] => '\<tau> Boolean" 
		            ("_ \<OclIncludesAll>'(_')" [66,65]65)

    "_OclExcludesAll_std":: "[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>::collection)VAL] => '\<tau> Boolean" 
		            ("_ \<OclExcludesAll>'(_')" [66,65]65)

    "_OclIsEmpty_std"    :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Boolean"
		            ("_ \<OclIsEmpty>'(')" [66])

    "_OclNotEmpty_std"   :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Boolean" 
		            ("_ \<OclNotEmpty>'(')" [66])

    "_OclIncluding_std"  :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
                            ("_ \<OclIncluding>'( _ ')") 

    "_OclExcluding_std"  :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
		            ("_ \<OclExcluding>'( _ ')") 

    "_OclComplement_std" :: "('\<tau>, '\<beta>::collection) VAL => ('\<tau>, '\<beta>) VAL" 
		            ("_ \<OclComplement>'(')") 

    "_OclUnion_std"      :: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL"
                            ("_ \<OclUnion>'( _ ')"          [66,65]65)

syntax  (* standard ascii syntax *)
    "_OclSize_ascii"      :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Integer" 
		            ("_ ->size'(')" [66])

    "_OclCount_ascii"     :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] => '\<tau> Integer" 
		            ("_ ->count'(_')" [66,65]65)

    "_OclIncludes_ascii"  :: "[ ('\<tau>, '\<beta>::collection) VAL,('\<tau>, '\<alpha>) VAL] => '\<tau> Boolean" 
		            ("_ ->includes'(_')" [66,65]65) 

    "_OclExcludes_ascii"  :: "[('\<tau>, '\<beta>::collection) VAL,('\<tau>, '\<alpha>::bot) VAL] => '\<tau> Boolean"
 		            ("_ ->excludes'(_')" [66,65]65)

    "_OclFlatten_ascii"   :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ ->flatten'(')" [66])

    "_OclSum_ascii"       :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Integer"
		            ("_ ->sum'(')" [66])

    "_OclAsSet_ascii"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ ->asSet'(')" [66])

    "_OclAsSequence_ascii":: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ ->asSequence'(')" [66])

    "_OclAsBag_ascii"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ ->asBag'(')" [66])
 
    "_OclAsOrderedSet_ascii":: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("_ ->asOrderedSet'(')" [66])

    "_OclIncludesAll_ascii" :: "[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>::collection)VAL] => '\<tau> Boolean" 
		            ("_ ->includesAll'(_')" [66,65]65)

    "_OclExcludesAll_ascii" ::"[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>::collection)VAL] => '\<tau> Boolean" 
		            ("_ ->excludesAll'(_')" [66,65]65)

    "_OclIsEmpty_ascii"     :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Boolean"
		            ("_ ->isEmpty'(')" [66])

    "_OclNotEmpty_ascii"    :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Boolean" 
		            ("_ ->notEmpty'(')" [66])

    "_OclIncluding_ascii"   :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
                            ("_ ->including'( _ ')") 

    "_OclExcluding_ascii"   :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
                            ("_ ->excluding'( _ ')") 

    "_OclComplement_ascii"  :: "('\<tau>, '\<beta>::collection) VAL => ('\<tau>, '\<beta>) VAL" 
		            ("_ ->complement'(')") 

    "_OclUnion_ascii"       :: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL"
                               ("_ ->union'( _ ')"          [66,65]65)

    "_OclIntersection_ascii":: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL" 
                            ("_ ->intersection'( _ ')"   [71,70]70)


syntax  (xsymbols)
    "_OclSize_math"      :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Integer" 
		            ("\<MathOclSizeLeft>_\<MathOclSizeRight>" [66])

    "_OclCount_math"     :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] => '\<tau> Integer" 
		            ("_ \<MathOclCount> _ " [66,65]65)

    "_OclIncludes_math"  :: "[('\<tau>, '\<alpha>) VAL, ('\<tau>, '\<beta>::collection) VAL] => '\<tau> Boolean" 
		            ("_ \<MathOclIncludes> _" [66,65]65) 

    "_OclExcludes_math"  :: "[('\<tau>, '\<alpha>::bot) VAL, ('\<tau>, '\<beta>::collection) VAL] => '\<tau> Boolean"
 		            ("_ \<MathOclExcludes> _" [66,65]65)

    "_OclFlatten_math"      :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("\<MathOclFlattenLeft>_\<MathOclFlattenRight>" [66])

    "_OclSum_math"       :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Integer"
		            ("\<MathOclSum> _" [66])

    "_OclAsSet_math"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("\<MathOclAsSet> _" [66])

    "_OclAsSequence_math":: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("\<MathOclAsSequence> _" [66])

    "_OclAsBag_math"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("\<MathOclAsBag> _" [66])
 
    "_OclAsOrderedSet_math":: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>, '\<beta>::collection) VAL" 
		            ("\<MathOclAsOrderedSet> _" [66])

    "_OclIncludesAll_math" :: "[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>::collection)VAL] => '\<tau> Boolean" 
		            ("_ \<MathOclIncludesAll>'(_')" [66,65]65)

    "_OclExcludesAll_math" ::"[('\<tau>,'\<alpha>::collection)VAL,('\<tau>,'\<alpha>::collection)VAL] => '\<tau> Boolean" 
		            ("_ \<MathOclExcludesAll>'(_')" [66,65]65)

    "_OclIsEmpty_math"     :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Boolean"
		            ("\<MathOclIsEmpty>_" [66])

    "_OclNotEmpty_math"    :: "('\<tau>, '\<alpha>::collection) VAL => '\<tau> Boolean" 
		            ("\<MathOclNotEmpty>_" [66])

    "_OclIncluding_math"   :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
                           ("_ \<MathOclIncluding> _ ") 

    "_OclExcluding_math"   :: "[('\<tau>,'\<beta>::collection)VAL,('\<tau>,'\<alpha>::bot) VAL] \<Rightarrow> ('\<tau>,'\<beta>) VAL"
		           ("_ \<MathOclExcluding> _ ") 

    "_OclComplement_math"  :: "('\<tau>, '\<beta>::collection) VAL => ('\<tau>, '\<beta>) VAL" 
		           ("_ \<MathOclComplement>") 

    "_OclUnion_math"       :: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL"
                           ("_ \<MathOclUnion> _ "          [66,65]65)


(*<*)
parse_translation {*
  let
            
  in [("_OclIncludes_ascii", binop ascii "OclIncludes"),
      ("_OclIncludes_std",   binop std   "OclIncludes"),
      ("_OclIncludes_math",  flip_args(binop  math "OclIncludes")),
      ("_OclExcludes_ascii", binop ascii "OclExcludes"),
      ("_OclExcludes_std",   binop std   "OclExcludes"),
      ("_OclExcludes_math",  flip_args(binop  math "OclExcludes"))]
     @ 
     flat(map (emb binop) ["OclCount",
                           "OclIncluding","OclExcluding", 
                           "OclUnion",
                           "OclIncludesAll", "OclExcludesAll"]) @ 
     flat(map (emb monop) ["OclSize", "OclFlatten", "OclSum","OclAsSet",   
                           "OclAsSequence","OclAsBag","OclAsOrderedSet",
                           "OclComplement","OclIsEmpty", "OclNotEmpty"])
  end;
*}


print_translation {*
  let
  in [("OclIncludes", fn S => if !OCL_syntax_mode <> math 
                              then print_const "OclIncludes" S
                              else flip_args(print_const "OclIncludes") S),
      ("OclExcludes", fn S => if !OCL_syntax_mode <> math 
                              then print_const "OclExcludes" S
                              else flip_args(print_const "OclExcludes") S)]
     @
     map emb_print ["OclCount", "OclIncludesAll", "OclExcludesAll",
                    "OclSize", "OclFlatten", "OclSum","OclAsSet",   
                    "OclAsSequence","OclAsBag","OclAsOrderedSet",
                    "OclIsEmpty", "OclNotEmpty",
                    "OclIncluding","OclExcluding", 
                    "OclUnion", "OclComplement"]
  end;
*}

(*>*)
subsection {* Operations on Collection Types with Ordering *}
consts (* abstract collection operations *)
    OclPrepend  :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
    OclAppend   :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
    OclInsertAt :: "[('\<tau>,'\<alpha>::collection ) VAL,'\<tau> Integer,('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<alpha> )VAL"     
    OclAt       :: "[('\<tau>,'\<alpha>::collection ) VAL,'\<tau> Integer] => ('\<tau>, '\<beta>) VAL"
    OclIndexOf  :: "[('\<tau>,'\<alpha>::collection ) VAL,('\<tau>, '\<beta>) VAL] => '\<tau> Integer"
    OclFirst    :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>,'\<beta>) VAL"     
    OclLast     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>,'\<beta>) VAL"     
 
syntax  (* standard syntax *)
    "_OclPrepend_std"  :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
                           ("_ \<OclPrepend>'( _ ')" [66,65]65)
    "_OclAppend_std"   :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
                           ("_ \<OclAppend>'( _ ')" [66,65]65)
    "_OclInsertAt_std" :: "[('\<tau>,'\<alpha>::bot)Sequence,'\<tau> Integer,('\<tau>,'\<alpha>)VAL] => ('\<tau>,'\<alpha> )Sequence"     
                           ("_ \<OclInsertAt>'( _ , _ ')" [66,65,65]) 
    "_OclAt_std"       :: "[('\<tau>,'\<alpha>::collection ) VAL,'\<tau> Integer] => ('\<tau>, '\<beta>) VAL"
                           ("_ \<OclAt>'( _ ')" [66,65])
    "_OclIndexOf_std"  :: "[('\<tau>,'\<alpha>::collection ) VAL,('\<tau>, '\<beta>) VAL] => '\<tau> Integer"
                           ("_ \<OclIndexOf> '( _ ')" [66,65])
    "_OclFirst_std"    :: "('\<tau>, '\<alpha>::collection ) VAL => ('\<tau>,'\<beta>) VAL"     
                           ("_ \<OclFirst>" [66]) 
    "_OclLast_std"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>,'\<beta>) VAL"     
                           ("_ \<OclLast>" [66]) 

syntax  (* standard ascii syntax *)
    "_OclPrepend_ascii"  :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
                             ("_ ->prepend'( _ ')" [66,65]65)
    "_OclAppend_ascii"   :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
                             ("_ ->append'( _ ')" [66,65]65)
    "_OclInsertAt_ascii" :: "[('\<tau>,'\<alpha>::bot)Sequence,'\<tau> Integer,('\<tau>,'\<alpha>)VAL] => ('\<tau>,'\<alpha> )Sequence"
                             ("_ ->insertAt'( _ , _ ')" [66,65,65]) 
    "_OclAt_ascii"       :: "[('\<tau>,'\<alpha>::collection ) VAL,'\<tau> Integer] => ('\<tau>, '\<beta>) VAL"
                             ("_ ->at'( _ ')" [66,65])
    "_OclIndexOf_ascii"  :: "[('\<tau>,'\<alpha>::collection ) VAL,('\<tau>, '\<beta>) VAL] => '\<tau> Integer"
                             ("_ ->indexOf'( _ ')" [66,65])
    "_OclFirst_ascii"    :: "('\<tau>, '\<alpha>::collection ) VAL => ('\<tau>,'\<beta>) VAL"     
                             ("_ ->first'(')" [66]) 
    "_OclLast_ascii"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>,'\<beta>) VAL"     
                             ("_ ->last'(')" [66]) 

syntax  (xsymbols)
    "_OclPrepend_math"  :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
                            ("_ \<MathOclPrepend>  _ " [66,65]65)
    "_OclAppend_math"   :: "[('\<tau>,('\<alpha>::collection) ) VAL, ('\<tau>,'\<beta>::bot ) VAL] => ('\<tau>,'\<alpha> ) VAL"
                            ("_ \<MathOclAppend> _ " [66,65]65)
    "_OclInsertAt_math" :: "[('\<tau>,'\<alpha>::bot)Sequence,'\<tau> Integer,('\<tau>,'\<alpha>)VAL] => ('\<tau>,'\<alpha> )Sequence"
                            ("_ \<MathOclInsertAt> '( _ , _ ')" [66,65,65]) 
    "_OclAt_math"       :: "[('\<tau>,'\<alpha>::collection ) VAL,'\<tau> Integer] => ('\<tau>, '\<beta>) VAL"
                            ("_ \<MathOclAt> _ " [66,65])
    "_OclIndexOf_math"  :: "[('\<tau>,'\<alpha>::collection ) VAL,('\<tau>, '\<beta>) VAL] => '\<tau> Integer"
                            ("_ \<MathOclIndexOf> _ " [66,65])
    "_OclFirst_math"    :: "('\<tau>, '\<alpha>::collection ) VAL => ('\<tau>,'\<beta>) VAL"     
                            ("\<MathOclFirst> _" [66]) 
    "_OclLast_math"     :: "('\<tau>, '\<alpha>::collection) VAL => ('\<tau>,'\<beta>) VAL"     
                            ("\<MathOclLast> _ " [66]) 

(*
parse_translation {*
  let
            
  in  [
      ("_OclInsertAt_ascii",quant "OclInsertAt" ascii),
      ("_OclInsertAt_std",  quant "OclInsertAt" std),
      ("_OclInsertAt_math", flip_args(quant "OclInsertAt" math)),
      ]@
     flat(map (emb binop) ["OclPrepend",
                           "OclAppend", 
                           "OclAt", "OclIndexOf"]) @ 
     flat(map (emb monop) ["OclFirst", "OclLast"])
  end;
*}

print_translation {*
  let
  in map emb_print ["OclPrepend", "OclAppend", "OclInsertAt", "OclAt", "OclIndexOf",
	             "OclFirst", "OclLast"]
  end;
*}
*)

subsection {* Operations on Collection Types without Ordering *}
consts (* abstract collection operations *)
     OclIntersection:: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] \<Rightarrow> ('\<tau>,'\<beta>)VAL"
 
syntax  (* standard syntax *)
    "_OclIntersection_std":: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL" 
                             ("_ \<OclIntersection>'( _ ')"   [71,70]70)



syntax  (* standard ascii syntax *)
    "_OclIntersection_ascii":: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL" 
                            ("_ ->intersection'( _ ')"   [71,70]70)


syntax  (xsymbols)
    "_OclIntersection_math":: "[('\<tau>,'\<beta>::collection)VAL, ('\<tau>,'\<beta>)VAL] => ('\<tau>,'\<beta>)VAL" 
                           ("_ \<MathOclIntersection> _ "   [71,70]70)

(*<*)
parse_translation {*
  let
            
  in 
     flat(map (emb binop) ["OclIntersection"])
  end;
*}

print_translation {*
  let
  in map emb_print ["OclIntersection"]
  end;
*}
(*>*)

section{* The Syntax of Iterators *}
text{* 
  This section contains the operators of \OCL involving binding,
  which as subsumed under the category ``iterators'' in the standard.
  Since we use a shallow embedding, we heavily use \acf{hoas} here. 
  
  However, quantifiers were not defined on the basis of iterate in
  \holocl. This is since iterate is undefined for all infinite
  collections, in contrast to quantifiers, which do not necessarily
  have computational content.
*}

consts 

   OclForAll   ::"[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] => '\<tau> Boolean"

   OclSelect   ::"[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] 
                  => ('\<tau>,'\<beta>::collection)VAL"

   OclReject   ::"[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] 
                  => ('\<tau>,'\<beta>::collection)VAL"

   OclCollect  ::"[('\<tau>,'\<gamma>::collection) VAL, 
                   ('\<tau>,'\<alpha>::bot)VAL => ('\<tau>,'\<beta>::bot)VAL] 
                  => ('\<tau>,'\<delta>::collection)VAL"

   OclCollectNested  ::"[('\<tau>,'\<gamma>::collection) VAL, 
                         ('\<tau>,'\<alpha>::bot)VAL => ('\<tau>,'\<beta>::bot)VAL] 
                        => ('\<tau>,'\<delta>::collection)VAL"


   OclIterate  ::"[('\<tau>,'\<gamma>::collection) VAL, 
                   [('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<beta>::bot)VAL] => ('\<tau>,'\<beta>::bot)VAL,
                   ('\<tau>,'\<beta>::bot)VAL] 
                  => ('\<tau>,'\<beta>::bot)VAL"

   OclIsUnique ::"[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => ('\<tau>,'\<gamma>::bot)VAL] => '\<tau> Boolean"

   OclOne      ::"[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] => '\<tau> Boolean"

   OclAny      ::"[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] 
                 => ('\<tau>,'\<alpha>::bot)VAL"

constdefs
   OclExists   :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL => '\<tau> Boolean] => '\<tau> Boolean"
  "OclExists S P \<equiv>  \<MathOclNot>(OclForAll S (\<lambda> x. \<MathOclNot> (P x)))" 


syntax (* OCL standard *)
  "_OclForAll_std"  :: "('\<tau>,'\<beta>::collection) VAL \<Rightarrow> idt \<Rightarrow> '\<tau> Boolean" 
                       ("_ \<OclForAll>'(_ | _')" [71,100,70]50)

  "_OclExists_std"  :: "('\<tau>,'\<beta>::collection) VAL \<Rightarrow> idt \<Rightarrow> '\<tau> Boolean" 
                       ("_ \<OclExists>'(_ | _')" [71,100,70]50)

  "_OclSelect_std"  :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, '\<tau> Boolean] 
                       => ('\<tau>,'\<beta>::collection)VAL"
                       ("_ \<OclSelect>'(_ | _')" [71,100,70]50)

  "_OclReject_std"  :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, '\<tau> Boolean] 
                       => ('\<tau>,'\<beta>::collection)VAL"
                       ("_ \<OclReject>'(_ | _')" [71,100,70]50)
  
  "_OclCollect_std" :: "[('\<tau>,'\<gamma>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<beta>::bot)VAL] 
                       => ('\<tau>,'\<delta>::collection)VAL"
                       ("_ \<OclCollect>'(_ | _')" [71,100,70]50)

  "_OclCollectNested_std" :: "[('\<tau>,'\<gamma>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<beta>::bot)VAL] 
                             => ('\<tau>,'\<delta>::collection)VAL"
                             ("_ \<OclCollectNested>'(_ | _')" [71,100,70]50)

  "_OclIterate_std" :: "[('\<tau>,'\<beta>::collection) VAL, idt, idt, '\<gamma>, '\<delta>] 
                       => ('\<tau>,'\<delta>)VAL"
                       ("_ \<OclIterate>'(_;_=_ | _')" [71,100,70]50)

  "_OclOne_std"     :: "[('\<tau>,'\<beta>::collection) VAL, '\<alpha>, '\<tau> Boolean] => '\<tau> Boolean"
                       ("_ \<OclOne>'(_ | _')" [71,100,70]50)

  "_OclIsUnique_std":: "[('\<tau>,'\<beta>::collection) VAL, '\<alpha>, '\<tau> Boolean] => '\<tau> Boolean"
                       ("_ \<OclIsUnique>'(_ | _')" [71,100,70]50)



syntax 
  "_OclForAll_ascii" :: "('\<tau>,'\<beta>::collection) VAL \<Rightarrow> idt \<Rightarrow> '\<tau> Boolean" 
                        ("_ ->forAll'(_ | _')" [71,100,70]50)

  "_OclExists_ascii"  :: "('\<tau>,'\<beta>::collection) VAL \<Rightarrow> idt \<Rightarrow> '\<tau> Boolean" 
                        ("_ ->exists'(_ | _')" [71,100,70]50)

  "_OclSelect_ascii"  :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, '\<tau> Boolean] 
                         => ('\<tau>,'\<beta>::collection)VAL"
                        ("_ ->select'(_ | _')" [71,100,70]50)

  "_OclReject_ascii"  :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, '\<tau> Boolean] 
                         => ('\<tau>,'\<beta>::collection)VAL"
                        ("_ ->reject'(_ | _')" [71,100,70]50)
  
  "_OclCollect_ascii" :: "[('\<tau>,'\<gamma>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<beta>::bot)VAL] 
                        => ('\<tau>,'\<delta>::collection)VAL"
                        ("_ ->collect'(_ | _')" [71,100,70]50)

  "_OclCollectNested_ascii" :: "[('\<tau>,'\<gamma>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<beta>::bot)VAL] 
                              => ('\<tau>,'\<delta>::collection)VAL"
                              ("_ ->collectNested'(_ | _')" [71,100,70]50)

  "_OclIterate_ascii"  :: "[('\<tau>,'\<beta>::collection) VAL, idt, idt, '\<gamma>, '\<delta>] 
                        => ('\<tau>,'\<delta>)VAL"
                        ("_ ->iterate'(_;_=_ | _')" [71,100,70]50)

  "_OclOne_ascii"      ::  "[('\<tau>,'\<beta>::collection) VAL, '\<alpha>, '\<tau> Boolean] => '\<tau> Boolean"
                        ("_ ->one'(_ | _')" [71,100,70]50)

  "_OclIsUnique_ascii" :: "[('\<tau>,'\<beta>::collection) VAL, '\<alpha>, '\<tau> Boolean] => '\<tau> Boolean"
                        ("_ ->isUnique'(_ | _')" [71,100,70]50)


syntax (xsymbols)
  "_OclForAll_math" :: "idt \<Rightarrow> ('\<tau>,'\<beta>::collection) VAL \<Rightarrow> '\<tau> Boolean" 
                         ("\<MathOclForAll> _ \<in> _ \<bullet> _" [71,100,70]50)

  "_OclExists_math" :: "idt \<Rightarrow> ('\<tau>,'\<beta>::collection) VAL \<Rightarrow> '\<tau> Boolean" 
                         ("\<MathOclExists> _ \<in> _ \<bullet> _" [71,100,70]50)

  "_OclSelect_math"  :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, '\<tau> Boolean] 
                         => ('\<tau>,'\<beta>::collection)VAL"
                        ("\<MathOclSelectLeft>_ : _ | _\<MathOclSelectRight>" [71,100,70]50)

  "_OclReject_math"  :: "[('\<tau>,'\<beta>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, '\<tau> Boolean] 
                         => ('\<tau>,'\<beta>::collection)VAL"
                        ("\<MathOclRejectLeft>_ : _ | _\<MathOclRejectRight>" [71,100,70]50)
  
  "_OclCollect_math" :: "[('\<tau>,'\<gamma>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<beta>::bot)VAL] 
                        => ('\<tau>,'\<delta>::collection)VAL"
                        ("\<MathOclCollectLeft> _ \<MathOclIncludes> _ | _ \<MathOclCollectRight>"                         [71,100,70]50)

(* TODO: correct syntax for collectNested, should be:
                              ("\<MathOclCollectNestedLeft>  _ \<MathOclIncludes> _ | _ \<MathOclCollectNestedRight>"                         [71,100,70]50)
*)

  "_OclCollectNested_math" :: "[('\<tau>,'\<gamma>::collection) VAL, ('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<beta>::bot)VAL] 
                              => ('\<tau>,'\<delta>::collection)VAL"
                              ("\<MathOclCollectLeft>* _ \<MathOclIncludes> _ | _ *\<MathOclCollectRight>"                         [71,100,70]50)

  "_OclIterate_math" :: "[('\<tau>,'\<beta>::collection) VAL, idt, idt, '\<gamma>, '\<delta>] 
                        => ('\<tau>,'\<delta>)VAL"
                        ("_ \<MathOclIterate>'(_;_=_ | _')" [71,100,70]50)

  "_OclOne_math"     ::  "[('\<tau>,'\<beta>::collection) VAL, '\<alpha>, '\<tau> Boolean] => '\<tau> Boolean"
                        ("_ \<MathOclOne>'(_ | _')" [71,100,70]50)

  "_OclIsUnique_ascii":: "[('\<tau>,'\<beta>::collection) VAL, '\<alpha>, '\<tau> Boolean] => '\<tau> Boolean"
                        ("_ \<MathOclIsUnique>'(_ | _')" [71,100,70]50)


(*<*)
parse_translation {*
  let
     fun quant name mode [obj, Free(x,T), b] =
         (if !OCL_strict_syntax_checking andalso !OCL_syntax_mode <> mode 
          then error(String.concat["strict checking: No ",
                                   OCL_syntax2string mode,
                                   " syntax allowed here"])
          else Syntax.const name $ obj $ Term.absfree (x, T, b))

     fun iter mode [obj, Free(x1,T1), Free(x2,T2), init, b] =
            (if !OCL_strict_syntax_checking andalso !OCL_syntax_mode <> mode 
             then error(String.concat["strict checking: No ",
                                   OCL_syntax2string mode,
                                   " syntax allowed here"])
             else Syntax.const "OclIterate" 
                 $ obj 
                 $ Term.absfree (x1, T1, Term.absfree (x2, T2,b))
                 $ init)

            
  in [("_OclForAll_ascii", quant "OclForAll" ascii),
      ("_OclForAll_std",   quant "OclForAll" std),
      ("_OclForAll_math",  flip_args(quant "OclForAll" math)),
      ("_OclExists_ascii", quant "OclExists" ascii),
      ("_OclExists_std",   quant "OclExists" std),
      ("_OclExists_math",  flip_args(quant "OclExists" math)),
      ("_OclSelect_ascii", quant "OclSelect" ascii),
      ("_OclSelect_std",   quant "OclSelect" std),
      ("_OclSelect_math",  flip_args(quant "OclSelect" math)),
      ("_OclReject_ascii", quant "OclReject" ascii),
      ("_OclReject_std",   quant "OclReject" std),
      ("_OclReject_math",  flip_args(quant "OclReject" math)),
      ("_OclCollect_ascii",quant "OclCollect" ascii),
      ("_OclCollect_std",  quant "OclCollect" std),
      ("_OclCollect_math", flip_args(quant "OclCollect" math)),
      ("_OclCollectNested_ascii",quant "OclCollectNested" ascii),
      ("_OclCollectNested_std",  quant "OclCollectNested" std),
      ("_OclCollectNested_math", flip_args(quant "OclCollectNested" math)),
      ("_OclOne_ascii",    quant "OclOne" ascii),
      ("_OclOne_std",      quant "OclOne" std),
      ("_OclOne_math",     flip_args(quant "OclOne" math)),
      ("_OclIsUnique_ascii",quant"OclIsUnique" ascii),
      ("_OclIsUnique_std", quant "OclIsUnique" std),
      ("_OclIsUnique_math",flip_args(quant "OclIsUnique" math)),
      ("_OclIterate_ascii",iter ascii),
      ("_OclIterate_std",  iter std),
      ("_OclIterate_math", iter math)
      ]
  end;
*}

print_translation {*
  let fun quant_trans name (obj:: Abs abs :: R) = 
          let val mode    = !OCL_syntax_mode
              val tag     = String.concat["_",name,"_",(OCL_syntax2string mode)]
              val consthd = Const(tag,dummyT)
              val (x,T)   = atomic_abs_tr' abs
          in  if mode = math 
              then list_comb(consthd, x :: obj :: T ::R)
              else list_comb(consthd, obj :: x :: T ::R)
          end 

       fun iter_trans name (obj:: Abs(x1,T1,Abs(x2,T2, b)):: init::R) =
          let val mode    = !OCL_syntax_mode
              val tag     = String.concat["_",name,"_",(OCL_syntax2string mode)]
              val consthd = Const(tag,dummyT)
              val (X1,Abs(_,_,T))  = atomic_abs_tr' (x1,T1,Abs(x2,T2, b))
              val (X2,T') = atomic_abs_tr' (x2,T2, T)             
          in  list_comb(consthd, obj :: X1 :: X2 :: init :: T' :: R) 
          end

  in
      [("OclForAll",   quant_trans "OclForAll"),
       ("OclExists",   quant_trans "OclExists"),
       ("OclSelect",   quant_trans "OclSelect"),
       ("OclReject",   quant_trans "OclReject"),
       ("OclCollect",  quant_trans "OclCollect"),
       ("OclCollectNested",  quant_trans "OclCollectNested"),
       ("OclOne",      quant_trans "OclExists"),
       ("OclIsUnique", quant_trans "OclIsUnique"),
       ("OclIterate",  iter_trans  "OclIterate")
      ]
  end
*}

(*>*)

section{* Semantics: Derived Concepts *}
text{* 
  In principle, the following concepts could be defined---more or less
  directly as according their specification in the
  standard---conservatively as constant definition, similarly to an
  abbreviation.  However, we refrained from this possibility due to
  our theory-morpher technique, which requires as many as possible
  operator definitions in combinator format and derives the resulting
  properties from there. Consequently, we stated the requirements
  following from the standard simply as unproven lemmas and prove them
  later. *}

text{* The following lemmas are unproven--they mirror the requirements in the
       standard and are intended as typechecked equivalents. These collection
       are proven on the basis of the concrete definitions in the concrete
       collection class theories.

       One might wonder why the Isabelle-concept of an axiomatic class
       is not used here. The reason is that the underlying type
       concept of Isabelle (Order-Sorted Polymorphism in Isabelle) is
       too weak to express the connection between the type
       constructors to be specified and their content; in the
       subsequent declaration for example, there is no connection
       between @{text "'\<beta>::collection"} and text @{text "'\<alpha>::bot"} as
       there should be. These types in the abstract class are
       therefore quite rough approximations. As a consequence,
       axiomatic classes would easily require very general properties
       that can not be met by any concrete instance. 
*}


subsection{* Count *}
text{* The standard requires the following definition: *}
(*<*)
ML{*OCL_strict_syntax_checking := false *};
(*>*)
lemma REQ_11_7_1_4: 
  "self ->count(obj)  \<equiv> 
  self ->iterate(elem;acc=0 | if elem \<MathOclStrongEq> obj then acc + 1 else acc endif)" 
  oops

text{* Unfortunately, for the case of sets, this definition assumes
       that sets are ``duplicate free lists'' which is simply not
       the case in general, in particular if we admit infinite sets
       as in \holocl. *}

text{* Instead, we define the properties pointwise for each collection. *}

subsection{* Excludes *}

text{* The standard defines exclusion via \inlineocl+ ->count +, which is possible,
       but messy with hindsight to a standard set theory: *}
 
    lemma REQ11_7_1_3:     "self ->excludes(obj) \<equiv> (self ->count(obj) \<MathOclStrongEq> 0)" oops

subsection{* includesAll *}

text{* The definition of \inlineocl+ ->includesAll + strictly follows the standard
       REQ11\_7\_1\_5: *}

    lemma REQ11_7_1_5: 
    "(self ->includesAll(obj)) 
    \<equiv> (obj ->forAll(elem | (self ->includes(elem))))" oops

text{* Unfortunately, this form of definition is \emph{not} conservative 
       for type technical reasons: elem is too polymorphic and contains 
       a free type variable not occuring in the lhs of this equation.
       Solution: \inlineocl+ ->includesAll + must be defined individually
       on each concrete collection. *}

text{* With the following three definitions, the situation is the same. *} 

    lemma REQ11_7_1_6:
    "(self ->excludesAll(obj)) \<equiv> (obj ->forAll(elem | (self ->excludes(elem))))" 
    oops


    lemma REQ11_7_1_7: "isEmpty self \<equiv> (self ->size() \<MathOclStrongEq> 0)" oops
 

    lemma REQ11_7_1_8: "notEmpty X  \<equiv> not(isEmpty(X))" oops
end
