(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_OrderedSet.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL OrderedSet *}

theory OCL_OrderedSet 
imports 
  "$HOLOCL_HOME/src/library/collection/OCL_Collection" 
begin
 text_raw{* \label{thy:OCL_OrderedSet} *}

constdefs 

 (* collection operations *)
  
    sosetsize     :: "('a, 'b::bot OrderedSet_0) VAL => 'a Integer"
                                                ("_ ->size'( ')" [66])
   "sosetsize     \<equiv> lift1(strictify(\<lambda>X.
                                \<HolOclLiftLeft> int(size(drop (Rep_OrderedSet_0 X))) \<HolOclLiftRight>))"


    OCLincludes     :: "[('a, 'b::bot) VAL, ('a, 'b OrderedSet_0) VAL] => 'a Boolean" 
		                               ("_ ->includes'( _ ')" [66,65]65) 
   "OCLincludes     \<equiv> lift2(strictify(\<lambda>x. 
                             strictify(\<lambda>X. 
                                \<HolOclLiftLeft> x mem drop (Rep_OrderedSet_0 X) \<HolOclLiftRight> )))" 

   excludes     :: "[('a, 'b::bot) VAL, ('a, 'b OrderedSet_0) VAL] => 'a Boolean" 
		                               ("_ ->excludes'( _ ')" [66,65]65) 
  "excludes     \<equiv> lift2(strictify(\<lambda>x.
                             strictify(\<lambda>X. 
                                 \<HolOclLiftLeft> ~(x mem drop(Rep_OrderedSet_0 X)) \<HolOclLiftRight> )))"

  sosetcount    :: "[('a, 'b::bot OrderedSet_0) VAL, ('a, 'b) VAL] => 'a Integer"
                                               ("_ ->count'( _ ')" [66,65]65)
   "sosetcount   \<equiv> lift2(strictify(\<lambda>X.
                             strictify(\<lambda>x.
                                \<HolOclLiftLeft> (int(size((List.filter (\<lambda>e. (x = e)) (drop(Rep_OrderedSet_0 X)))))) \<HolOclLiftRight>)))"
(*    
   sosetincludesall     :: "[('a, 'b::bot OrderedSet_0) VAL, ('a, 'b OrderedSet_0) VAL] => 'a Boolean"
                                               ("_ ->includesall'( _ ')" [66,65]65)
   "sosetincludesall     \<equiv> lift2(strictify(\<lambda>X.
                             strictify(\<lambda>Y.
                                \<HolOclLiftLeft> drop(Rep_OrderedSet_0 Y) <=  drop (Rep_OrderedSet_0 X)\<HolOclLiftRight>)))"
 *)
   sosetexcludesall     :: "[('a, 'b::bot OrderedSet_0) VAL, ('a, 'b OrderedSet_0) VAL] => 'a Boolean"
                                               ("_ ->excludesall'( _ ')" [66,65]65)
  "sosetexcludesall     \<equiv> lift2(strictify(\<lambda>X.
                             strictify(\<lambda>Y.
                                 \<HolOclLiftLeft> ((set(drop(Rep_OrderedSet_0 X)) Int set(drop(Rep_OrderedSet_0 Y))) = {} ) \<HolOclLiftRight>)))"
  
    sosetisempty     :: "('a, 'b::bot OrderedSet_0) VAL => 'a Boolean"
                                                ("_ ->isEmpty'( ')" [66])
   "sosetisempty     \<equiv> lift1(strictify(\<lambda>X.
                                \<HolOclLiftLeft> ([] = (drop (Rep_OrderedSet_0 X))) \<HolOclLiftRight>))"
 
    sosetnotempty     :: "('a, 'b::bot OrderedSet_0) VAL => 'a Boolean"
                                                ("_ ->notEmpty'( ')" [66])
   "sosetnotempty    \<equiv> lift1(strictify(\<lambda>X.
                                \<HolOclLiftLeft> ~([] = (drop (Rep_OrderedSet_0 X))) \<HolOclLiftRight>))"
 
 





(* smashed order set  Operations *)

 
   insert'      :: "[('a,'b) VAL,('a,('b::bot) OrderedSet_0) VAL] => ('a,'b OrderedSet_0) VAL"
  "insert'       \<equiv> (lift2(strictify(\<lambda>X.
                           strictify(\<lambda>Y. 
                             Abs_OrderedSet_0(lift (remdups (concat[ [X],(drop(Rep_OrderedSet_0 Y))])))))))"

   union        :: "[('a,('b::bot) OrderedSet_0) VAL, ('a,'b OrderedSet_0) VAL] => ('a,'b OrderedSet_0) VAL"
                                               ("_ ->union'( _ ')" [66,65]65)
  "union        \<equiv> (lift2(strictify(\<lambda>X.
                           strictify(\<lambda>Y. 
                             (Abs_OrderedSet_0(lift
                              (remdups (concat[(drop(Rep_OrderedSet_0 X)),(drop(Rep_OrderedSet_0 Y))]))))))))"
end



