(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_OrderedSet.thy -
 * Copyright (C) 2005,2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL OrderedSet *}
theory OCL_OrderedSet 
imports 
  "$HOLOCL_HOME/src/library/collection/smashed/OCL_Sequence" 
begin
 text_raw{* \label{thy:OCL_OrderedSet} *}

section{* Properties of the Congruence inside the Datatype Adaption *}

text{* The following rules are transformed versions of the automatically
       generated rules for datatype conversions. *}
lemma Abs_OrderedSet_inject_charn:
  "\<lbrakk> (\<bottom>::'a::bot) \<notin> set x; (\<bottom>::'a::bot) \<notin> set y; distinct x; distinct y \<rbrakk> \<Longrightarrow> 
   (((Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_OrderedSet_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_OrderedSet_0_inject, 
     simp_all add: OrderedSet_0_def smash_def)
(* TODO: which form is to be preferred *)
(* as an elimination rule:
lemma Abs_OrderedSet_inject_absurd1:
  "\<lbrakk> (Abs_OrderedSet_0 down) = (Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>); (\<bottom>::'a::bot) \<notin> set x \<rbrakk> \<Longrightarrow> P"
  by(subst (asm) Abs_OrderedSet_0_inject,
     simp_all add: OCL_OrderedSet_type.OrderedSet_0_def smash_def)
  or as a simplifier rule:

  - are both of them needed?
*)

lemma Abs_OrderedSet_0_inject_absurd11 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x; distinct x \<rbrakk> \<Longrightarrow> ((Abs_OrderedSet_0 down) = (Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_OrderedSet_0_inject,
     simp_all add: OrderedSet_0_def smash_def)

lemma Abs_OrderedSet_0_inject_absurd12 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x; distinct x \<rbrakk> \<Longrightarrow> ((Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_OrderedSet_0 down)) = False"
  by(subst Abs_OrderedSet_0_inject,
     simp_all add: OrderedSet_0_def smash_def)

lemma Abs_OrderedSet_0_inject_absurd21 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x; distinct x \<rbrakk> \<Longrightarrow> ( \<bottom> = (Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_OrderedSet_def)

lemma Abs_OrderedSet_0_inject_absurd22 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x; distinct x \<rbrakk> \<Longrightarrow> ((Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_OrderedSet_def)

lemma Abs_OrderedSet_0_inject_charn:
  "\<lbrakk> (\<bottom>::'a::bot) \<notin> set x; (\<bottom>::'a::bot) \<notin> set y; distinct x; distinct y \<rbrakk> \<Longrightarrow> 
   (((Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_OrderedSet_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_OrderedSet_0_inject, 
     simp_all add: OrderedSet_0_def smash_def)

lemma Abs_OrderedSet_0_inverse_charn1 [simp]:
  "\<lbrakk> (\<bottom> \<notin> set x); distinct x \<rbrakk> \<Longrightarrow> Rep_OrderedSet_0 (Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_OrderedSet_0_inverse, 
      simp_all add: OrderedSet_0_def smash_def)

lemma Abs_OrderedSet_0_inverse_charn2 [simp]:
  "\<lbrakk> (\<bottom> \<notin> set x); distinct x \<rbrakk> \<Longrightarrow> \<HolOclDropLeft>Rep_OrderedSet_0 (Abs_OrderedSet_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_OrderedSet_0_inverse, 
      simp_all add: OrderedSet_0_def smash_def)

lemma Abs_OrderedSet_0_cases_charn:
  assumes bottomCase : "\<lbrakk> x = \<bottom> \<rbrakk> \<Longrightarrow> P"
  assumes listCase   : "\<And>y. (\<lbrakk> x = Abs_OrderedSet_0 \<HolOclLiftLeft>y\<HolOclLiftRight>; \<bottom> \<notin> set y; distinct y \<rbrakk> \<Longrightarrow> P)"
  shows                "P"
  apply(rule_tac x="x" in Abs_OrderedSet_0_cases)
  apply(case_tac "y = \<bottom>")
  apply(rule bottomCase, simp add: UU_OrderedSet_def)
  apply(frule not_down_exists_lift2[THEN iffD1])
  apply(rule_tac y="\<HolOclDropLeft>y\<HolOclDropRight>" in listCase)
  apply(auto simp: OrderedSet_0_def smash_def)
  apply(case_tac "distinct ya", simp_all)
  done

lemma Abs_OrderedSet_0_induct_charn:
  assumes bottomCase : "P \<bottom>"
  assumes stepCase   : "\<And>y. (\<lbrakk> \<bottom> \<notin> set y; distinct y \<rbrakk> \<Longrightarrow> P (Abs_OrderedSet_0 \<HolOclLiftLeft>y\<HolOclLiftRight>))"
  shows                "P x"
  apply(rule_tac x="x" in Abs_OrderedSet_0_induct)
  apply(rule_tac  x="Abs_OrderedSet_0 y" in Abs_OrderedSet_0_cases_charn)
  apply(simp_all add: bottomCase stepCase)
  done

lemma inj_on_Abs_OrderedSet: "inj_on Abs_OrderedSet_0 OrderedSet_0";
  by(rule inj_on_inverseI,rule OrderedSet_0.Abs_OrderedSet_0_inverse,assumption)

lemma inj_Rep_OrderedSet : "inj Rep_OrderedSet_0"
  by(rule inj_on_inverseI, rule OrderedSet_0.Rep_OrderedSet_0_inverse)

lemma smashed_orderedSet_charn: 
  "(\<bottom> \<notin> set X \<and> distinct X) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> OrderedSet_0)"
  by(unfold smash_def OrderedSet_0_def UU_OrderedSet_def,auto) (*characterization 1 *)

lemma UU_in_smashed_orderedSet [simp]: 
  "\<bottom> \<in> OrderedSet_0"
  by(unfold smash_def OrderedSet_0_def UU_OrderedSet_def,auto) (*characterization 2a *)

lemma down_in_smashed_orderedSet [simp]: 
  "down \<in> OrderedSet_0"
  by(unfold smash_def OrderedSet_0_def UU_OrderedSet_def,auto) (*characterization 2b *)

lemma mt_in_smashed_orderedSet[simp]: 
  "(\<HolOclLiftLeft>[]\<HolOclLiftRight>) \<in> OrderedSet_0"
  by (unfold smash_def OrderedSet_0_def,auto)

lemma DEF_Abs_OrderedSet: "\<And>X. \<lbrakk>\<bottom> \<notin> set X; distinct X \<rbrakk> \<Longrightarrow> DEF (Abs_OrderedSet_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  apply(unfold DEF_def UU_OrderedSet_def, simp)
  done

lemma DEF_Rep_OrderedSet: 
  "\<And>X. DEF X \<Longrightarrow> DEF (Rep_OrderedSet_0 X)"
  apply(unfold DEF_def UU_OrderedSet_def,auto)
  apply(drule_tac f = "Abs_OrderedSet_0" in arg_cong)
  apply(simp add: Rep_OrderedSet_0_inverse)
  done

lemma not_DEF_Rep_OrderedSet: 
  "\<And>X. \<not> DEF X \<Longrightarrow> \<not> DEF (Rep_OrderedSet_0 X)"
  apply(unfold DEF_def UU_OrderedSet_def,auto)
  apply(simp add: Abs_OrderedSet_0_inverse)
  done


lemma exists_lift_OrderedSet: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. Rep_OrderedSet_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  by (drule DEF_Rep_OrderedSet,simp add: DEF_X_up)

lemma exists_lift_OrderedSet2: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. \<bottom> \<notin> set c \<and> distinct c \<and> Rep_OrderedSet_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  apply(frule exists_lift_OrderedSet)
  apply(clarify, simp)
  apply(subgoal_tac "\<HolOclLiftLeft>c\<HolOclLiftRight> \<in> OrderedSet_0")
  apply(erule smashed_orderedSet_charn[THEN iffD2])
  apply(drule sym, simp add: Rep_OrderedSet_0)
  done

lemma Rep_OrderedSet_cases: 
  "Rep_OrderedSet_0 X = \<bottom> \<or> (\<exists>c. \<bottom> \<notin> set c \<and> distinct c \<and> Rep_OrderedSet_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(case_tac "DEF X")
  apply(drule exists_lift_OrderedSet2)
  apply(drule_tac [2] not_DEF_Rep_OrderedSet)
  apply(simp_all add: DEF_def)
  done


lemma DEF_X_OrderedSet_0 : "DEF(X) = (\<exists>c. \<bottom> \<notin> set c \<and> distinct c \<and> Rep_OrderedSet_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(insert  Rep_OrderedSet_cases [of "X"], auto)
  apply(drule exists_lift_OrderedSet,auto)
  apply(rule swap)
  prefer 2
  apply(rule not_DEF_Rep_OrderedSet, auto)
  done


text{* The following lemma is crucial for the theory morpher: *} 
lemma DEF_X_OrderedSet: "DEF X = (\<exists>c.  \<bottom> \<notin> set c \<and> distinct c \<and> X = Abs_OrderedSet_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply (auto simp: DEF_Abs_OrderedSet)
  apply (simp add:DEF_X_OrderedSet_0)
  apply (erule exE, rule exI, auto)
  apply (rule injD [OF inj_Rep_OrderedSet])
  apply (auto simp: smashed_orderedSet_charn Abs_OrderedSet_0_inverse)
  done

text{* This lemma is very convenient in simplifing unfolded definitions *}
lemma DEF_X_OrderedSet':
  "\<lbrakk> X \<noteq> \<bottom> \<rbrakk>  \<Longrightarrow> 
   (\<exists>c. (\<bottom> \<notin> set c) \<and> (distinct c) \<and> (Rep_OrderedSet_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>)) \<and> 
   (\<exists>c. (\<bottom> \<notin> set c) \<and> (distinct c) \<and> (X = (Abs_OrderedSet_0 \<HolOclLiftLeft>c\<HolOclLiftRight>)))"
  apply(fold DEF_def)
  apply(frule DEF_X_OrderedSet[THEN iffD1])
  apply(drule DEF_X_OrderedSet_0[THEN iffD1])
  apply(simp)
  done
 
(* TODO: does this hurt to have both versions in the assumptions? *)
lemma OrderedSet_sem_cases_0:
  assumes defC: "\<And> c d. \<lbrakk> X \<noteq> \<bottom>; \<bottom> \<noteq> X; 
                          (\<bottom> \<notin> set c); (\<bottom> \<notin> set d); 
                          distinct c; distinct d;
                          Rep_OrderedSet_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          X = Abs_OrderedSet_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X"
  and   undefC: "\<lbrakk> X = \<bottom> \<rbrakk> \<Longrightarrow> P X"
  shows "P X"
  apply(rule Abs_OrderedSet_0_cases_charn, erule undefC)
  apply(rule defC, auto)
  done

lemma OrderedSet_sem_cases:
  assumes defC: "\<And> c d. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (\<bottom> \<notin> set c); (\<bottom> \<notin> set d); 
                          distinct c; distinct d;
                          Rep_OrderedSet_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_OrderedSet_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and   undefC: "\<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and      cpP: "cp P"
  shows "P X \<tau>"
  apply(rule_tac P2="P" in subst[OF sym[OF cp_subst]], rule cpP)
  apply(rule OrderedSet_sem_cases_0)
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule defC) prefer 9
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule undefC, simp_all)
  done

lemma OrderedSet_sem_cases_ext:
  assumes defC: "\<And> c d \<tau>. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (\<bottom> \<notin> set c); (\<bottom> \<notin> set d); 
                          distinct c; distinct d;
                          Rep_OrderedSet_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_OrderedSet_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and   undefC: "\<And> \<tau>. \<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and      cpP: "cp P"
  and      cpQ: "cp Q"
  shows "P X = Q X"
  apply(rule ext)
  apply(rule_tac X="X" in OrderedSet_sem_cases)
  apply(rule defC) prefer 9
  apply(rule undefC)
  apply(simp_all add: cpP cpQ)
  done

text{* These four lemmas are used by \inlineholocl{ocl_setup_op}
        to reason about definedness: *}

lemma lift2_strict_is_isdef_fw_OrderedSet_Val:
  assumes  f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_OrderedSet_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_OrderedSet_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g1: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; distinct a; (\<bottom>::'a::bot) \<noteq> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  and     inv_g2: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; distinct a; (\<bottom>::'a::bot) \<noteq> b \<rbrakk> \<Longrightarrow> distinct (g a b))"
  shows           " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule ext)
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def ss_lifting')
  apply(rule_tac X="X x" in OrderedSet_sem_cases_0)
  apply(rule impI, frule_tac b="(Y x)" in  inv_g1)
  apply(drule_tac[3] b="(Y x)" in inv_g2)
  apply(simp_all add: neq_commute)
  done

lemma lift2_strict_is_isdef_fw_OrderedSet_OrderedSet:
  assumes  f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_OrderedSet_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_OrderedSet_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_OrderedSet_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g1: "!!a b. \<lbrakk> \<bottom> \<notin> set a; distinct a; \<bottom> \<notin> set b; distinct b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b)"
  and     inv_g2: "!!a b. \<lbrakk> \<bottom> \<notin> set a; distinct a; \<bottom> \<notin> set b; distinct b \<rbrakk> \<Longrightarrow> distinct (g a b)"
  shows           " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule ext)
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def ss_lifting')
  apply(rule_tac X="X x" in OrderedSet_sem_cases_0)
  apply(rule_tac X="Y x" in OrderedSet_sem_cases_0)
  apply(rule impI, frule_tac b="ca" in  inv_g1)
  apply(drule_tac[4] b="ca" in inv_g2)
  apply(simp_all add: neq_commute)
  done

lemma lift2_strictify_implies_LocalValid_defined_OrderedSet_Val:
  assumes  f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_OrderedSet_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_OrderedSet_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g1: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; (\<bottom>::'a::bot) \<noteq> b\<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  and     inv_g2: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; (\<bottom>::'a::bot) \<noteq> b\<rbrakk> \<Longrightarrow> distinct (g a b))"
  shows           " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_OrderedSet_Val)
  apply(assumption)+
  apply(simp_all add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done


lemma lift2_strictify_implies_LocalValid_defined_OrderedSet_OrderedSet:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_OrderedSet_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_OrderedSet_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_OrderedSet_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g1: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; distinct a; \<bottom> \<notin> set b; distinct b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  and     inv_g2: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; distinct a; \<bottom> \<notin> set b; distinct b \<rbrakk> \<Longrightarrow> distinct (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_OrderedSet_OrderedSet)
  apply(assumption)+
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

section{* Building a canonic representation of orderedSets *}

subsection{* The empty orderedSet *}

constdefs
    OclMtOrderedSet   :: "('a, 'b::bot OrderedSet_0) VAL"    
   "OclMtOrderedSet  \<equiv> lift0(Abs_OrderedSet_0(\<HolOclLiftLeft>[]\<HolOclLiftRight>))"

    syntax  (* standard syntax *)
    "_OclMtOrderedSet_std"   :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<OclMtOrderedSet>")
    syntax  (* ascii syntax *)
    "_OclMtOrderedSet_ascii" :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("mtOrderedSet")
    syntax  (xsymbols)
    "_OclMtOrderedSet_math"  :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<MathOclMtOrderedSet>")
    parse_translation {* flat(map (emb trans_const) ["OclMtOrderedSet"]) *}
    print_translation {* map emb_print ["OclMtOrderedSet"] *}


lemma OCL_is_isdef_OclMtOrderedSet [simp]:
  "\<MathOclValid> \<MathOclIsDefined> \<MathOclMtOrderedSet>"
  by(simp add: OclValid_def OclIsDefined_def OclTrue_def 
               OclMtOrderedSet_def ss_lifting')

lemma OCL_is_defopt_OclMtOrderedSet [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> \<MathOclMtOrderedSet>"
  by(simp add: valid_elim)



subsection{* And the 'cons' operation: including *}

defs
   (* including  :: "[('a,('b::bot) ) OrderedSet, ('a,'b::bot ) VAL] => ('a,'b ) OrderedSet" *)
   OclIncluding_def   :
   "OclIncluding  \<equiv> lift2(strictify(\<lambda>S. strictify (\<lambda>e. Abs_OrderedSet_0
                       (\<HolOclLiftLeft>remdups (concat [\<HolOclDropLeft>Rep_OrderedSet_0 S\<HolOclDropRight>, [e] ])\<HolOclLiftRight>))))"

(* TODO: correct setup op to handle this case as well *)
ocl_setup_op [OclIncluding]



lemma OCL_is_def_OclIncluding:
  "\<MathOclIsDefined>(OclIncluding (X::('a,'b::bot)OrderedSet) (Y::('a,'b::bot)VAL)) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  by(rule lift2_strict_is_isdef_fw_OrderedSet_Val[OF OclIncluding_def], simp_all)
  
lemma OCL_is_defopt_OclIncluding:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclIncluding (X::('a,'b::bot)OrderedSet) (Y::('a,'b::bot)VAL))) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  by(rule lift2_strictify_implies_LocalValid_defined_OrderedSet_Val[OF OclIncluding_def], simp_all)



subsection{* The syntax translation mkOrderedSet builds now our orderedSets *}

syntax
   "@OclFinOrderedSet" :: "args \<Rightarrow> ('a,'b OrderedSet_0)VAL"           ("mkOrderedSet[(_)]")

translations
  "mkOrderedSet[x, xs]"    == "OclIncluding (mkOrderedSet[xs]) x"
  "mkOrderedSet[x]"        == "OclIncluding OclMtOrderedSet x"

lemma test : "mkOrderedSet[1,2,3,4,5] = ?X" oops

subsection{* Relation of mtOrderedSet and including *}


text{* These two theorems should actually suffice to decide inequality of orderedSets *}
(* TODO: name for not strong equaltiy  . . . and representation *)
lemma including_notstrongeq_mtOrderedSet[simp]:
   "(\<MathOclNot> ((X::('a,'b::bot)OrderedSet) \<MathOclIncluding> (a::('a,'b)VAL) \<MathOclStrongEq> \<MathOclMtOrderedSet>)) = \<MathOclTrue>"
  apply(rule_tac X="X" in OrderedSet_sem_cases_ext)
  apply(simp_all add: cp_strongEq)
  apply(simp_all add: OclIncluding_def OclMtOrderedSet_def OclStrongEq_def
                      OclTrue_def OclNot_def localValid2sem ss_lifting' 
                      Abs_OrderedSet_inject_charn neq_commute)
  done


(* TODO: name for not strict equality\<dots> and representation *)
lemma including_notstricteq_mtOrderedSet[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('a,'b)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('a,'b::bot)OrderedSet) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> (X \<MathOclIncluding> a) `<>` \<MathOclMtOrderedSet>"
  apply(rule_tac X="X" in OrderedSet_sem_cases)
  apply(simp_all add: cp_strongEq cp_strictEq localValidDefined2sem)
  apply(simp_all add: OclIncluding_def OclMtOrderedSet_def OclStrictEq_def
                      OclTrue_def OclNot_def localValid2sem ss_lifting' 
                      Abs_OrderedSet_inject_charn neq_commute)
  done


section{* A second orderedSet constructor on the \OCL level *}
defs
   (* collectionRange :: "[('a,'b::{bot,plus}) VAL,('a,'b) VAL] \<Rightarrow> ('a,'b) OrderedSet" *)
   OclCollectionRange_def :
   "OclCollectionRange \<equiv> lift2 (strictify (\<lambda> x\<Colon>Integer_0. strictify (\<lambda> y. 
                           Abs_OrderedSet_0 \<HolOclLiftLeft>map (\<lambda> z::nat. \<HolOclLiftLeft>(int z) + \<HolOclDropLeft>x\<HolOclDropRight>\<HolOclLiftRight>) 
                                               [0..<(nat (\<HolOclDropLeft>y\<HolOclDropRight> - \<HolOclDropLeft>x\<HolOclDropRight> + 1))]\<HolOclLiftRight>)))"

ocl_setup_op [OclCollectionRange]

(* TODO: 1. add further rules to simplify use of CollectionRange
            e.g. includes and definedness
         2. is there an induction rule over OCL Integers? *)

lemma OCL_is_defopt_OclCollectionRange [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> ((OclCollectionRange (a::('a Integer)) b)::('a, Integer_0)OrderedSet) = 
   ((\<tau> \<MathOclValid> \<MathOclIsDefined> a) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> b))"
  oops


lemma collectionRange_mtOrderedSet_conv:
  "\<lbrakk> \<tau> \<MathOclValid> (b::'a Integer) \<MathOclLess> (a::'a Integer) \<rbrakk> \<Longrightarrow>
   OclCollectionRange a b \<tau> = (\<MathOclMtOrderedSet>::('a,Integer_0)OrderedSet) \<tau>"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> a", ocl_hypsubst, simp)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> b", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(simp_all add: DEF_def OclCollectionRange_def OclMtOrderedSet_def 
                      OclLess_def localValid2sem ss_lifting')
  done

lemma collectionRange_singleton_UC[simp]:
  "OclCollectionRange a a = \<MathOclMtOrderedSet>::('a,Integer_0)OrderedSet \<MathOclIncluding> (a::('a,Integer_0)VAL)"
  apply(rule ext, case_tac "x \<MathOclValid> \<MathOclIsUndefined>(a)", ocl_hypsubst, simp)
  apply(simp add: OclCollectionRange_def OclMtOrderedSet_def OclIncluding_def
                  localValidDefined2sem ss_lifting')
  done

lemma collectionRange_expand_including:
  "\<lbrakk> \<tau> \<MathOclValid> (a::'a Integer) \<MathOclLe> b \<rbrakk> \<Longrightarrow>
   ((OclCollectionRange a b)::('a,Integer_0)OrderedSet) \<tau> = 
   (OclCollectionRange a (b - 1)) \<MathOclIncluding> b \<tau>"
  oops


section{* The conversion operators *}

subsection{* OrderedSet to sequence *}
defs
   (* asSequnece     :: "('a,'b::bot) OrderedSet \<Rightarrow> ('a,'b::bot) Sequence" *)
   OclAsSequence_def : 
   "OclAsSequence    \<equiv> lift1 (strictify (\<lambda>X. 
                            Abs_Sequence_0 (Rep_OrderedSet_0 X)))"

ocl_setup_op [OclAsSequence]

subsection{* OrderedSet to bag *}
defs
   (* asBag     :: "('a,'b::bot) OrderedSet \<Rightarrow> ('a,'b::bot) Bag" *)
   OclAsBag_def  : 
   "OclAsBag     \<equiv> lift1 (strictify (\<lambda>X. 
                          Abs_Bag_0 \<HolOclLiftLeft>multiset_of \<HolOclDropLeft>Rep_OrderedSet_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

ocl_setup_op [OclAsBag]


subsection{* OrderedSet to orderedSet *}
defs
   (* asOrderedSet      :: "('a,'b::bot) OrderedSet \<Rightarrow> ('a,'b::bot) OrderedSet" *)
   OclAsOrderedSet_def  : 
   "OclAsOrderedSet     \<equiv> id"

lemma asOrderedSet_identity[simp]:
  "OclAsOrderedSet (S::('a,'b::bot)OrderedSet) = S"
  by(simp add: OclAsOrderedSet_def)



subsection{* OrderedSet to set *}
defs
   (* asSet     :: "('a,'b::bot) OrderedSet \<Rightarrow> ('a,'b::bot) Set" *)
   OclAsSet_def : 
   "OclAsSet    \<equiv> lift1 (strictify (\<lambda>X. Abs_Set_0 \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_OrderedSet_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

ocl_setup_op [OclAsSet]

(* TODO: add the rest *)
end

