(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * invariant_encoder.sml - 
 * Copyright (C) 2006  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

signature CLASSIC_INVARIANT_ENCODER =
sig
    val add_invariants_1 : Rep_Core.Classifier -> typ -> Rep_Core.Classifier list -> theory -> theory 			 
    val add_invariants_2 : Rep_Core.Classifier -> typ -> Rep_Core.Classifier list -> theory -> theory 			 
end  

structure classic_invariant_encoder : CLASSIC_INVARIANT_ENCODER =
struct

open Rep_Core
open holocl_term
open rep_encoder_common
     
open Rep_OclType
open Rep_OclTerm
     
open Rep_HolOclTerm
     
     
fun inv_of  thy cl u C  = map (fn (n,c) => (n, Ocl2holocl l1 thy cl u c)) (invariant_of C) 
			  
				    
fun mkSndLevelOCL selftype u thy lhs defs prevLevel undef = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
				 Const("==", dummyT) 
				      $ (lhs)
				      $ Term.absfree("self",selftype, 
						     OclIf u (defs (* $ Free("x",dummyT) *)) 
							   (prevLevel $ Free("self",selftype) )
							undef
				 )],propT))

fun add_invariants_1 C u cl  thy =
    let
	val thy_name = Context.theory_name thy

	val name = name_of C	
	val _ = trace 3  ("        "^(string_of_path name))
	val invariants  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => (rm_whitespace s, snd c)
							   | NONE => ("inv_"^Int.toString(n), snd c )
					   )) (0,(inv_of thy cl u C))) 


	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 


        (* add INV constant *)
	val inv_type      = mk_funT(V(u,class_u),BOOLT(St(u)))  

	val thy = foldl (fn (constraint,thy) 
			    =>  add_consts_i (SOME (hc_path l2 (invariant_1_of  (name_of C))))
				    [((fst constraint),inv_type,NoSyn)] thy) 
		  thy invariants

	 val invs = map (fn (n,t) =>  (n,#1(Sign.infer_types (Sign.pp  thy)  thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", inv_type --> inv_type --> propT) $
                               Const ((string_of_path ((hc_path l2 (invariant_1_of  (name_of C)))))^"."^n,inv_type) $ 
			       Term.absfree ("self",V(u,class_u), t)
			      )
			      ],propT)))) invariants 

	val thy = foldl (fn (constraint,thy) 
			    =>  add_definitions_i false (SOME ((hc_path l2 (invariant_1_of  (name_of C)))))
				    [((fst constraint)^"_def",(snd constraint))] thy) 
		  thy invs 

	(* add top-level invariant *)
	val thy = add_consts_i (SOME ((hc_path l2 (invariant_1_of  (name_of C))))) 
		  [((local_name_of (hc_path l2 (invariant_1_of  (name_of C) ))),inv_type,NoSyn)] thy 



	  

	val c_str_i = map (fn s => Const((string_of_path  (hc_path l2 (invariant_1_of  (name_of C))))
					 ^"."^(fst s),dummyT)$Free("obj",dummyT)) invariants


	val inv_def_i =  (#1(Sign.infer_types (Sign.pp thy)  thy
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", inv_type --> inv_type --> propT) 
				      $ (Const(string_of_path (hc_path l2 (invariant_1_of  (name_of C))),dummyT)) 
				      $ (Term.absfree ("obj",dummyT,(foldr1 (fn (a,b) => OclAnd u a b) c_str_i))))
			       ],propT)))


	val thy = add_definitions_i false  (SOME ((hc_path l2 (invariant_1_of  (name_of C)))))  
				  [((local_name_of (hc_path l2 (invariant_1_of  (name_of C)))),inv_def_i)] thy 



     in 
	thy
    end


fun add_invariants_2 C u cl  thy =
    let
	fun collect_path u oclterm =  let
	in 
	    case oclterm of 
		Literal _                  => []
	      | CollectionLiteral _        => []
	      | If (cterm,ctype, tterm,ttype,eterm,etype,iftype) => List.concat [(collect_path u cterm),  
										 (collect_path u tterm), 
										 (collect_path u eterm)]
	      (**************************************)
	      (* AssociationEndCall                 *)
	      (**************************************)
	      | AssociationEndCall(src,ts,p,t)    => [ Ocl2holocl l1 thy cl u (AssociationEndCall(src,ts,p,t))]  
	      (**************************************)
	      (* AttributeCall                      *)
	      (**************************************)
	      | AttributeCall(src,ts,p,t)         => [ (Ocl2holocl l1 thy cl u (AttributeCall(src,ts,p,t)))] 
	      (**************************************)
	      (* OperationCall                      *)
	      (**************************************)
	      | OperationCall (src,st,opn,args,rt) => let 
		    val argsdefs = List.concat (map (fn (a,at) => collect_path u a) args)
		in
		    List.concat [collect_path u src, argsdefs] 
		end
	      (**************************************)
	      (* Variable                           *) 
	      (**************************************)
	      | Variable (vname,t) => [Ocl2holocl l1 thy  cl u ( Variable (vname,t))] 
	      (**************************************)
	      (* Let                                *)
	      (**************************************) 
	      (* Error                              *)
	      | Let (var,vtyp,rhs,rhstyp,interm,intyp) => List.concat [(collect_path u rhs), 
      								       (collect_path u interm)]
	      (**************************************)
	      (* OperationWithType                  *)
	      (**************************************)
	      (* Error                              *)
	      | OperationWithType (src,_,_,_,_) =>collect_path u src 
	      (**************************************) 
	      (* Iterate                            *)
	      (**************************************)
	      (* Error                              *)
	      | Iterate (vars,rvar,rvtyp,rterm,sterm,styp,iterm,ityp,rtyp) =>  
		error ("error: unknown Iterate '"^(rvar)^"' in in ocl2holocl") 
	      (**************************************)
	      (* Iterator                           *)
	      (**************************************)
	      (* forAll                              *)
	      | Iterator (_,_,src,_,c,_,_) => [] (*  List.concat [[Ocl2holocl l1 thy cl u src ],(collect_path u src), 
      							    (collect_path u c)] *)
	      (**************************************)
	      (* Catch out                          *)
	      (**************************************)
	      (* Error                              *)
	      | _ => error ("error: unknown OCL-term in in collect_path") 
	end
				     
	val thy_name = Context.theory_name thy
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))

	val invariants  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => (rm_whitespace s, snd c)
							   | NONE => ("inv_"^Int.toString(n), snd c )
					   )) (0,(inv_of thy cl u C))) 

	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
        val class_u = Type((string_of_path (hc_path univ (name_of C))),holes) 

        (* add INV constant *)
	val inv_type      = mk_funT(V(u,class_u),BOOLT(St(u)))  

	val thy = foldl (fn (constraint, thy) 
			    =>  add_consts_i (SOME (hc_path l2 (invariant_2_of  (name_of C))))
				    [((fst constraint),inv_type,NoSyn)] thy) 
		  thy invariants


	val invs = [] 
	val defs = case (invariant_of C) of
		       [] => []
		     | invariants => map (fn (n,c) =>   
					     let 
						 fun mk_unique [] = []
						   | mk_unique (x::xs) = if x mem xs then mk_unique xs else x::(mk_unique xs)

						 val l = (map (fn t => Const("OCL_Boolean.OclIsDefined",dummyT)$t) (collect_path u c) )
						 val l = mk_unique l
					     in
						 case l of 
						     [] => (n,OclTrue(u)) 
						   | _  => (n,foldr1 (fn (a,b) => OclAnd u a b) l) 
					     end
					 ) 
					 (invariants)

	val defs  = snd (foldl_map (fn (n,c) => (n+1,
							 case (fst c) of 
							     (SOME(s)) => (rm_whitespace s, snd c)
							   | NONE => ("inv_"^Int.toString(n), snd c )
					   )) (0,defs)) 



	 val invs = map (fn (n,def) =>  (n,#1(Sign.infer_types (Sign.pp  thy)  thy
				 (K NONE) (K NONE) [] true ([
						  mkSndLevelOCL (V(u,class_u)) u thy 	
								(Const ((string_of_path ((hc_path l2 (invariant_2_of  (name_of C)))))^"."^n,inv_type))
								def
								(Const ((string_of_path ((hc_path l2 (invariant_1_of  (name_of C)))))^"."^n,inv_type))
						   		(Const("OCL_Undefined.OclUndefined",dummyT))
				  
			      ],propT)))) defs





      
	val thy = foldl (fn (constraint,thy) 
			    =>  add_definitions_i false (SOME ((hc_path l2 (invariant_2_of  (name_of C)))))
				    [((fst constraint)^"_def",(snd constraint))] thy) 
		  thy invs 


	(* val thy = add_consts_i (SOME (hc_path l2 (invariant_2_of  (name_of C))))  *)
	val thy = add_consts_i (SOME (hc_path l2 ( (name_of C)))) 
			       [(((local_name_of (hc_path l2 (invariant_2_of  (name_of C))))),
				 inv_type,NoSyn)] thy 

	val c_str_i = map (fn s => Const((string_of_path  (hc_path l2 (invariant_2_of  (name_of C))))^"."^(fst s),
					 dummyT)$Free("obj",dummyT)) invariants


	val inv_def_i =  (#1(Sign.infer_types (Sign.pp thy) thy 
				 (K NONE) (K NONE) [] true ([
			      (Const  ("==", inv_type --> inv_type --> propT) 
				      $ (Const(string_of_path (hc_path l2 (invariant_2_of  (name_of C))),dummyT)) 
				      $ (Term.absfree ("obj",dummyT,(foldr1 (fn (a,b) => OclAnd u a b) c_str_i))))
			       ],propT)))

	val thy = add_definitions_i false  (SOME ((hc_path l2 ( (name_of C)))))  
				  [((local_name_of (hc_path l2 (invariant_2_of  (name_of C)))),inv_def_i)] thy 
    in 
	thy
    end
end
 
