(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * mdr_encoder_level0.sml - generic mdr import interface for HOL-OCL
 * Copyright (C) 2001-2005  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


signature REP_ENCODER_LEVEL0 =
sig
    val build_universe : typ -> Rep_Core.Classifier list -> theory -> (theory * typ)
    val  add_level0    : typ -> Rep_Core.Classifier list -> theory -> theory
end


structure rep_encoder_level0 : REP_ENCODER_LEVEL0 =
struct

open Rep_OclType;
open Rep_Core;
open rep_encoder_common

fun ext_DEF_tac thy = 
    let
	val ctx = thy
	fun ext_DEF_1_tac thy topthm = 
	    let
		val  next = case hd(prems_of(topthm)) of
				_ $ (_ $ (_ $ (foo $ _))) $ _  => foo
			      | _  => Const("dummy",dummyT) 
	    in
		(case next of
		     Const("Lifting.FromR",_)     
		     => ((forward_tac [get_thm (thy) (Name "DEF_FromRD")] 1) 
			     THEN (rotate_tac ~1 1))
		   | Const("Lifting.FromL",_)     
		     => ((forward_tac [get_thm (thy) (Name "DEF_FromLD")] 1) 
			     THEN (rotate_tac ~1 1))
		   | Const("OCL_Universe.base",_) 
		     => ((forward_tac [get_thm (thy) (Name "DEF_baseD")] 1) 
			     THEN (rotate_tac ~1 1))
		   | _ => no_tac			
		) topthm
	    end 
    in
	(((forward_tac [get_thm thy (Name "DEF_FromLD")] 1) 
	      ORELSE (forward_tac [get_thm thy (Name "DEF_FromRD")] 1) 
	      ORELSE (forward_tac [get_thm thy (Name "DEF_baseD")] 1)
	      ORELSE (no_tac)) 
	     THEN (rotate_tac ~1 1)
	     THEN (REPEAT (ext_DEF_1_tac thy)))
    end



fun hat Integer              = IntegerT
  | hat String               = StringT
  | hat Real                 = RealT
  | hat Boolean              = BooleanT
  | hat (Set(t))             = SetT (hat t)
  | hat (OrderedSet(t))      = OrderedSetT (hat t)
  | hat (Bag(t))             = BagT (hat t)
  | hat (Sequence(t))        = SequenceT (hat t)
  | hat (Classifier(s))      = RefT
  | hat _                    = error "Error in hat()"
  | hat DummyT              = dummyT
			     
			      
(*
 * simple type substitution 
 *)
fun subst_type_tfree (from, to) t = 
    let 
        fun subst (s, t) = if String.compare(from,s) = EQUAL 
			   then to 
			   else TFree(s,t);
    in   
	map_type_tfree subst t
    end  
    
(*    
 fun stripProdT (Type("*",[_,R])) = stripProdT R
   | stripProdT x =  x
 *)	    
fun stripSumT (Type("+",[_,R])) = stripSumT R
  | stripSumT x =  x
		   
fun insert_lu_in_universe u lu C    = let	       
    fun get_tfree_of_next_son  (Type("*",[_,Type("Lifting.up",[TFree(X,typeS)])])) = X
      | get_tfree_of_next_son  (Type("*",[_,Type("Lifting.up",[S])]))= 
	let 
	    val TFree(X,typeS) = stripSumT S 
	in 
	    X 
	end
      | get_tfree_of_next_son _ = error("Error in get_tfree_of_next_son")
				  
    val luC = case (get_type_universe u C)  of 
		  SOME(tu) => tu 
		| _  =>  error ("Error: cannot get type universe ("^(string_of_path C)^")")
    val next = get_tfree_of_next_son (dest_LiftT luC)
in
    subst_type_tfree (next,lu) u
end
				      
				      
fun alpha C = "'a_"^(short_name_of C) 
fun beta  C = "'b_"^(short_parent_name_of C) 
	      
fun alpha_type C = TFree(alpha C,typeS)
fun beta_type  C = TFree(beta C,typeS)
		   
		   
fun cl_ext thy_name  C = let
    val base_name  = string_of_path (hc_path univ (base_type_name_of (name_of C)))
    val Ctype = Type(base_name,[])
in
    cl_extT(alpha_type C,Ctype)
end
			
fun insert_classifier_in_universe u thy_name  C = let
    fun cl_u thy_name  C = let
	val base_name  = string_of_path (hc_path univ (base_type_name_of (name_of C)))
	val Ctype = Type(base_name,[])
    in
	extT(alpha_type C,beta_type C,Ctype)
    end
    val lu = cl_u thy_name C
in
    insert_lu_in_universe u lu ( (hc_path univ (base_type_name_of (parent_name_of C))))
end
	

(* 
 * generate datatype for class identifier 
 *)

fun add_classifier_key thy C  =
    let
 	val name = short_name_of C
	val key  = local_name_of (key_name_of (name_of C))
	val pkg_name = (string_of_path o package_of) C 
	val _    = trace 9  (" Generating ID for classifier: "
			     ^((string_of_path o name_of) C)^" ==> "
			     ^(string_of_path(hc_path univ (name_of C))))
	val (oldpath,thy) = set_thy_path (SOME (hc_path univ (package_of C))) thy
	val (thy, {distinct, inject, exhaustion, rec_thms,
		   case_thms, split_thms, induction, size, simps}) =
	    DatatypePackage.add_datatype false
					 [key] [([], key, NoSyn, 
					       [(name, [], NoSyn)])] thy
	val thy = Theory.add_types [] thy 

	val (_,thy)  = set_thy_path oldpath thy
    in
	thy
    end
    
fun add_classifier_type thy  C  =
    let		   
	fun base_type_of thy C  =
	    let
		val key  =  string_of_path (hc_path univ (key_name_of (name_of C)))
		val a_types    = map (fn a => hat (#attr_type a)) (attributes_of C)
		val base_type  = case a_types of 
				     [] => unitT 
				   | l => foldr1 mk_prodT l
	    in
		mk_prodT(Type(key,[]),base_type)
	    end
	    
	fun c_type_of thy  C   =
	    let
		val pkg_name = (string_of_path o package_of) C
		val thy_name = Context.theory_name  thy
		val base_type  = base_type_of thy C
	    in
		if ((parent_name_of C) = (name_of OclAnyC))
		then base_type	
		else mk_prodT(Type((string_of_path(
				    hc_path univ
					    (base_type_name_of (parent_name_of C)))) ,[]), 
			      base_type) 
	    end
	val thypath = (hc_path univ (package_of C))		      
	val thy_name = Context.theory_name thy
	val name = name_of C
	val base_type = local_name_of (base_type_name_of (name_of C))
	val C_type  =local_name_of (type_name_of (name_of C))
	val pkg_name = package_of C
	val thy = add_tyabbrs_i (SOME thypath) [(base_type, [], 
						 base_type_of thy C,
						 NoSyn)] thy		  
		  
	val thy = add_tyabbrs_i (SOME thypath) [(C_type, [], 
						 c_type_of thy C,
						 NoSyn)] thy
		  
        val C_ext = (Type((string_of_path( 
			   hc_path univ  (ext_name_of (parent_name_of C))),
			   [cl_ext thy_name C])))


        val thy = add_tyabbrs_i (SOME thypath) [(local_name_of (ext_name_of name), [alpha C], 
				 C_ext, NoSyn) ] thy
    in
	thy 
    end




fun add_type_of_universe  u thy = 
    let
	val thy_name = Context.theory_name thy
	val uname = "U_"^thy_name
	val thy = add_tyabbrs_i NONE [(uname, collect_holes u, u, NoSyn)]
				thy
    (* \todo{add syntax translation} 
     *)
    (*  val thy = Theory.add_trrules Syntax.PrintRule ((uname, u), 
						       ("type", ))] thy *)
    in
	thy 
    end

    

fun add_class_type thy u C  =
    let
	val thy_name = Context.theory_name thy
	val name = short_name_of C
	val parent = short_parent_name_of C
	val pkey =  ((hc_path univ (base_type_name_of (name_of C))))
	val ext = the(get_type_universe u pkey)
	val ctype = Type( (string_of_path (hc_path univ (ext_name_of (parent_name_of C)))),[dest_LiftT(ext)])
			      
val ctype = LiftT(ctype)

	val thypath = (hc_path univ (package_of C))

	val thy = add_tyabbrs_i (SOME thypath ) [(name, collect_holes ext, ctype, 
					 NoSyn)] thy
    (* \todo{add syntax translation} *) 
     (* 
     val thy = Theory.add_trrules Syntax.PrintRule (("logic", type), 
						    ("type", )) thy 
     *)
    in
	thy 
    end

fun n_times_s 0 s = []
  | n_times_s n s = (n_times_s (n-1) s)@[s]

fun n_times_ss 0 s e = []
  | n_times_ss 1 s e = [e] 
  | n_times_ss n s e = [s]@(n_times_ss (n-1) s e)
		    
   
fun pos_in_tsum thy_name  u C  =
    let
	val key  = hc_path univ (base_type_name_of (name_of C))
	val pkey = hc_path univ (base_type_name_of (parent_name_of C))

	val base_type  = local_name_of key
	val C_type  = local_name_of (type_name_of (name_of C))

	val p_class_u = case (get_type_universe u ( pkey)) of
			    SOME(u) => u
			  | NONE  => error("cannot get type universe of " 
					   ^ (string_of_path pkey))
 
	fun get_t_sum (Type("Lifting.up",[T]))  = get_t_sum T
	  | get_t_sum (Type("*",[_,T])) 	= get_t_sum T 
 	  | get_t_sum T = T         

	val t_sum = get_t_sum p_class_u		    

	fun count_pos_in_t_sum sum key = let
	    fun simp_sum  (Type("Lifting.up",[T]))  = simp_sum T  
	      | simp_sum   (Type("*",[T,T2])) = T
	      | simp_sum   (Type("+",[T1,T2])) = (Type("+",[simp_sum T1, 
							    simp_sum T2]))
	      | simp_sum T = T
	    val s_sum = simp_sum sum
	    fun count (Type("+",[(Type(s,_)),T2])) key =
		if (s=key) then 0 else  (count T2 key) + 1
	      | count _ _ =  error ("Error in pos_in_tsum")
	in
	    count s_sum key
	end				 
    in
	count_pos_in_t_sum t_sum (string_of_path key)  
    end
    
    
    
    
	       
	    	       
fun n_supclasses name cl = 
    if (name = (name_of OclAnyC)) 
    then 0
    else (n_supclasses (parent_name_of (class_of name cl )) cl) + 1  
  
fun add_attributes_0 C u  cl  thy =
    let
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))
	val thy_name = Context.theory_name thy
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)
	val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
	val base = implode((n_times_s ((n_supclasses name cl ) ) " o base"))
	fun pos_in_attr  a (x::xs) = if (a=x) then 0 
				     else (1 + (pos_in_attr a xs))
	  | pos_in_attr  a [] = 0
				
	fun add_consts thy_path thy  (a:attribute)  = 
	    let
		val attr_name    = #name a
		val attr      = #attr_type a

		val attr_type = mk_funT(class_u, hat (attr))
		val attr = local_name_of (attr_of name attr_name)
	    in
		add_consts_i (SOME thy_path) [(attr,attr_type,NoSyn)] thy
	    end
	    
	fun add_defs thy_path thy (a:attribute)  =
	    let
		val attr_n    = #name a
		val attr      = #attr_type a

		val attr_name = local_name_of (attr_of name attr_n)
		val attr_def_name = attr_name^"_def"
		val attrC = attributes_of C
		val sel_pair = if ((length attrC) > 1)
			       then implode ((n_times_s 
						  ((pos_in_attr a
							attrC) 
						  ) " snd o "))
			       else ""
		val sel_attr = if ((length attrC) =1) 
			       then "" 
			       else (if (a = (hd (rev attrC)))
				     then sel_pair
				     else "fst o "^sel_pair )
				    
		val attr_def = attr_name^" == "^sel_attr
			       ^"snd o fst o Lifting.drop"^base
		val _ = trace 9 (attr_def_name^": "^attr_def^"  ")
	    in
		add_definitions (SOME thy_path) [(attr_def_name,attr_def)] thy 	
	    end 
	val thy = foldl (fn (a,thy) => add_consts ((hc_path l0 (name_of C))) thy a) 
			thy (attributes_of C)
	val thy = foldl (fn (a,thy) => add_defs   ((hc_path l0 (name_of C))) thy a) 
			thy (attributes_of C)  
    in 
	thy
    end



    
fun add_constructors_0 C u cl thy =
    let
	val name = name_of C
	val _ = trace 3  ("        "^(string_of_path name))
	val thy_name = Context.theory_name thy
	val parent = parent_name_of C
	val mk = hc_path l0 ((mk_class_of o name_of) C)
	val mk_parent = hc_path l0 ((mk_class_of o parent_name_of) C)
	val get = hc_path l0 ((get_class_of o name_of) C)
	val get_parent = hc_path l0 ( (get_class_of o parent_name_of)  C)
	val is = hc_path l0 ((is_class_of o name_of) C)
	val isType = hc_path l0 ((is_classType_of o name_of) C)

	val is_parent = hc_path l0 ((is_class_of o parent_name_of) C)
	val is_univ = hc_path l0 ((is_univ_class_of o name_of) C)
	val is_univ_parent = hc_path l0 ((is_univ_class_of o parent_name_of) C) 

        val class_holes_str = (holes_of_type u C)
	val parent_holes_str = if (parent_name_of C = (name_of OclAnyC)) 
			       then []
			       else (holes_of_type 
					  u 
					 (class_of (parent_name_of C) cl)) 
	val univ_holes_str  = (holes_of_type  u OclAnyC)
	val univ_holes_unit_str = map (fn a => if (listin a parent_holes_str)
					       then a 
					       else "unit") univ_holes_str 
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type u C)

        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 

	val holes_par = map (fn x => TFree(x,typeS)) 
			    (holes_of_typestr  u (parent_name_of C))
        val class_u_par = if (parent_name_of C) = (name_of OclAnyC) then 
			      the(get_type_universe u (key_name_of (name_of OclAnyC))) (* TODO *)
			  else Type((string_of_path (hc_path univ (parent_name_of C))),holes_par)

	val class2parent = hc_path l0 (class2get_parent  name (local_name_of parent)) 
	val parent2class = hc_path l0 (parent2class_of  name (local_name_of parent))
	fun add_consts thy = 
	    let
		val mk_type      = mk_funT(class_u,u)
		val get_type     = mk_funT(u,class_u)  
		val is_type      = mk_funT(class_u,boolT)
		val is_univ_type = mk_funT(u,boolT)
		val class_type   = mk_setT(u)
		val class2parent_type   = mk_funT(class_u, class_u_par)
		val parent2class_type   = mk_funT(class_u_par, class_u)
	       in
		   add_consts_i (SOME (hc_path l0 (package_of C)) )
				[(local_name_of mk,mk_type,NoSyn), 
					(local_name_of get,get_type,NoSyn),
					(local_name_of is,is_type,NoSyn), 
					(local_name_of isType,is_type,NoSyn), 
					(local_name_of is_univ,is_univ_type,NoSyn), 
					(local_name_of name,class_type, NoSyn), 
					(local_name_of class2parent,class2parent_type, NoSyn),
					(local_name_of parent2class,parent2class_type, NoSyn)
				       ] thy
	    end

	fun add_defs C thy = 
	    let
		val mk_def_name    = (local_name_of o mk_def_of) mk
	        val get_def_name   = (local_name_of o mk_def_of) get
	        val is_def_name    = local_name_of (mk_def_of is)
		val isType_def_name    = local_name_of (mk_def_of isType)
	        val is_univ_def_name = local_name_of (mk_def_of is_univ)
	        val class_def_name = local_name_of (mk_def_of name)
       		val parent_name =  (short_parent_name_of C)
		val class2parent_def_name = local_name_of (mk_def_of class2parent)
		val parent2class_def_name = local_name_of (mk_def_of parent2class)
		val n_Inr = pos_in_tsum thy_name u C
	        val is_def         = (local_name_of is)^" == Lifting.DEF" 
				     ^ implode(n_times_s (n_supclasses name cl) 
							 " o OCL_Universe.base")
				       

	        val isType_def         = (local_name_of isType)^" == %obj. ("^(local_name_of is)
					 ^" obj) & ~ ((Lifting.DEF o OCL_Universe.base" 
					 ^ implode(n_times_s (n_supclasses name cl) 
							     " o OCL_Universe.base")
					 ^ " ) obj)"
		val class_def   =  (local_name_of name)^" == Collect("^(string_of_path is_univ)^")"
		val is_univ_def = (local_name_of is_univ)^" univ == ("^(string_of_path is_univ_parent)
				  ^" univ) & ((Lifting.DEF o Lifting.FromL  "
				  ^ implode((n_times_s (pos_in_tsum thy_name u C ) " o Lifting.FromR"))
				  ^ implode(n_times_s (n_supclasses name cl) 
						      " o OCL_Universe.base")
				  ^ " o "^(string_of_path get_parent)^") univ )"
		fun const_get get depth = if (depth = 0)
					  then  
					      "((Lifting.FromL  "
					      ^ implode((n_times_s (pos_in_tsum thy_name u C ) " o Lifting.FromR"))
					      ^ implode(n_times_s (n_supclasses name cl) 
								  " o OCL_Universe.base")
					      ^ " o "^get^") univ )"
					  else
					      "Lifting.lift( (OCL_Universe.sup"
					      ^ (implode ((n_times_s ((n_supclasses name cl)-depth)) 
							      " o OCL_Universe.base"))
					      ^" o "^get^") univ, "^ (const_get get (depth-1))^")"

		val get_def   = (local_name_of get)^" univ == "^(const_get (string_of_path get_parent) (n_supclasses name cl))

		fun const_mk depth = if (depth = 0) 
				     then  
					 "((Lifting.lift "
					 ^ implode((n_times_s (pos_in_tsum thy_name  u C ) " o Inr"))
					 ^ " o Inl o Lifting.drop"
					 ^  implode(n_times_s (n_supclasses name cl) " o OCL_Universe.base")
					 ^ ") obj )"
				     else
					 "Lifting.lift( (OCL_Universe.sup"
						 ^ implode(n_times_s ((n_supclasses name cl)-depth) " o OCL_Universe.base")
 				 ^") obj, "^(const_mk (depth-1))^")"

		val mk_def   = (local_name_of mk)^" obj == "^(string_of_path mk_parent)^"("^(const_mk (n_supclasses name cl))^")"
			       
		fun pr_list (x::[]) = x
		  | pr_list (x::xs) = x^", "^(pr_list xs)
		  | pr_list [] = ""	    
				 
				 
		val mk_type_str = "("^(pr_list class_holes_str)^")"^(string_of_path name)
				  ^" => ("^(pr_list univ_holes_unit_str)^")U_"^thy_name
		val mk_parent_type_str = "("^(pr_list parent_holes_str)^")"^(string_of_path parent)
					 ^" => ("^(pr_list univ_holes_unit_str)^")U_"^thy_name
					 
		val class2parent_def = (local_name_of class2parent)^" == "^
				       (if(parent_name_of C)= (name_of OclAnyC) 
					then (string_of_path get_parent)^" o "^(string_of_path mk)
					else (string_of_path get_parent)^" o ("
					    ^(string_of_path mk)^"::("^mk_type_str^"))" 
				       )
		val parent2class_def = (local_name_of parent2class)^" == "^
				       (if(parent_name_of C)= (name_of OclAnyC)
					then (string_of_path get)^" o "^(string_of_path mk_parent)
					else (string_of_path get)^" o ("^
					     (string_of_path mk_parent)^"::("^mk_parent_type_str^"))" 
				       )
	    in
		add_definitions (SOME (hc_path l0 (package_of C)))
		    [
                     (class_def_name,class_def),
		     (is_def_name,is_def),
		     (isType_def_name,isType_def) ,
		    (is_univ_def_name,is_univ_def)   , 
		     (get_def_name,get_def)  ,  
		    (mk_def_name,mk_def)  ,   
		     (class2parent_def_name,class2parent_def) ,   
		     (parent2class_def_name,parent2class_def) 
		   ] thy 	
	    end 

	val thy = add_consts  thy 
	val thy = add_defs  C thy     
     in 
	thy
    end
  
fun prove_0 C u  cl thy =
    let

	val sup_ss = HOL_ss addsimps [get_thm (theory "OCL") (Name("sup_def"))]
	val def_X_up_ss = HOL_ss addsimps [get_thm (theory "OCL") (Name("DEF_X_up"))]

	val thy_name = Context.theory_name thy
	val class  = name_of C
	val parent = parent_name_of C
	val _ = trace 3  ("        "^(string_of_path class))

	val ocl_ss = simpset_of (theory "OCL")
	val ocl_cla = claset_of (theory "OCL")   

	val (oldpath,thy) = set_thy_path (SOME (hc_path l0 (package_of C))) thy

 	(* is_mk_A_0: "is_A_0 obj ==> is_A_univ_0(mk_A_0 obj)" *)

	val is_parent = string_of_path (hc_path l0 ((is_class_of o parent_name_of) C))
	val get_parent = string_of_path (hc_path l0 ((get_class_of o parent_name_of) C))
	val mk_parent = string_of_path (hc_path l0 ((mk_class_of o parent_name_of) C))
	val is_parent_univ = string_of_path (hc_path l0 ((is_univ_class_of o parent_name_of) C))

	val is_parent_def  = get_thm thy (Name (is_parent^"_def"))
	val mk_parent_def  = get_thm thy (Name (mk_parent^"_def"))
	val get_parent_def = get_thm thy (Name (get_parent^"_def"))
	val is_parent_univ_def = get_thm thy (Name (is_parent_univ^"_def")) 

	val path = string_of_path (package_of C)

	val ppath = string_of_path (hc_path l0 (rev (tl (rev parent))))
 	val get_mk_parent =ppath^".get_mk_"^(local_name_of parent)^"_id"                     (* TODO *)
	val is_univ_get_parent = ppath^".is_"^(local_name_of parent)^"_univ_implies_is_get"  (* TODO *)
	val is_mk_parent =ppath^".is_mk_"^(local_name_of parent)                             (* TODO *) 
	val mk_get_parent = ppath^".mk_get_"^(local_name_of parent)^"_id"                    (* TODO *)

	val parent_simps =    (get_thm thy (Name(get_mk_parent)))   
			      ::(get_thm thy (Name(is_univ_get_parent))) 
			      ::(get_thm thy (Name(is_mk_parent))) 
			      ::(get_thm thy (Name(mk_get_parent))) 
			      ::[]

	fun mkGoal_cterm prems goal =  (cterm_of (sign_of thy) 
					     (#1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
						 Const ("==>",propT --> propT --> propT) $
						       (Const ("Trueprop",boolT --> propT) $ prems) $
						       (Const ("Trueprop",boolT --> propT) $ goal)
						 ],propT))))



	      
	fun is_mk_class class parent thy = 
	    let

		val name="is_mk_"^(local_name_of class)
		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_class_of class))),dummyT)$Free("obj",dummyT)) 
				      (Const(string_of_path (hc_path l0 (is_univ_class_of class)),dummyT)
					    $(Const(string_of_path (hc_path l0 (mk_class_of class)),dummyT)$Free("obj",dummyT)))
	        val _ = trace 9  ("proving "^name)
		val prems_str = (string_of_path (hc_path l0 (is_class_of class)))
				::(string_of_path (hc_path l0 (is_univ_class_of class)))
				::(string_of_path (hc_path l0 (mk_class_of class)))
				::(string_of_path (hc_path l0 (get_class_of class)))
				::[]
		val simps = is_parent_def::parent_simps
		val prems= map (fn def => get_def thy def) prems_str
		val thm = quick_and_dirty_prove_goalw_cterm thy prems goal_i
				      (fn p => [cut_facts_tac p 1, 
						simp_tac 
						    (ocl_ss addsimps (is_parent_def::parent_simps) ) 1 ] )
	    in
		(fst(PureThy.add_thms [((name,thm),[])] (thy)))
	    end    
	    
	    
	fun is_parent_mk_class class parent thy = 
	    let
		val name="is_"^(local_name_of parent)^"_mk_"^(local_name_of class)
			 
		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_class_of class))),dummyT)$Free("obj",dummyT)) 
				      (Const(is_parent_univ,dummyT)$(Const(string_of_path (hc_path l0 (mk_class_of class)),dummyT)
									  $Free("obj",dummyT)))
	        val _ = trace 9  ("proving "^name)

		val prems_str = (string_of_path (hc_path l0 (is_class_of class)))
				::(string_of_path (hc_path l0 (is_univ_class_of class)))
				::(string_of_path (hc_path l0 (mk_class_of class)))
				::(string_of_path (hc_path l0 (get_class_of class)))
				::[]

	
		val prems= map (fn def => get_def thy def) prems_str
		val thm = quick_and_dirty_prove_goalw_cterm thy prems goal_i
				      (fn p => [cut_facts_tac p 1, 
						simp_tac
						    (ocl_ss addsimps (is_parent_def::parent_simps)) 1 ])
	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end    
	    
	    
	    
	fun is_class_univ_implies_parent_univ class parent thy =
	    (*@ $is\_\template{C}\_univ_0 obj \Implies is\_A\_univ_0 obj$ @*)
	    let
		val name="is_"^(local_name_of class)^"_univ_implies_"^(local_name_of parent)^"_univ"
		val goal=(string_of_path (hc_path l0 (is_univ_class_of class)))^" obj ==> "^is_parent_univ^" obj"

		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_univ_class_of class))),dummyT)$Free("obj",dummyT)) 
				      (Const(is_parent_univ,dummyT)$Free("obj",dummyT))
	


	        val _ = trace 9  ("proving "^name)
		val simps =   is_parent_univ_def
			      ::(get_def thy (string_of_path (hc_path l0 (is_univ_class_of class))))
			      ::[]
		val thm = quick_and_dirty_prove_goalw_cterm thy [] goal_i
				      (fn prems => [cut_facts_tac prems 1,
						    ALLGOALS(asm_full_simp_tac
								 (ocl_ss addsimps (simps)))])
 	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end 
	fun is_class_univ_implies_is_get class parent thy =
	    (* is_D_univ_0 univ ==> is_D_0(get_D_0 univ) *)
	    let
		val name="is_"^(local_name_of class)^"_univ_implies_is_get"
		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_univ_class_of class))),dummyT)$Free("univ",dummyT)) 
				      (Const((string_of_path (hc_path l0 (is_class_of class))),dummyT)
					    $(Const(string_of_path (hc_path l0 (get_class_of class)),dummyT)$Free("univ",dummyT)))
	

		val _ = trace 9  ("proving "^name)
		val simps =   (get_def thy (string_of_path (hc_path l0 (is_univ_class_of class))))
			      ::(get_def thy (string_of_path (hc_path l0 (is_class_of class))))
			      ::(get_def thy (string_of_path (hc_path l0 (get_class_of class))))
			      ::[]
		val thm = quick_and_dirty_prove_goalw_cterm thy [] goal_i
				      (fn prems => [cut_facts_tac prems 1,
						    ALLGOALS(asm_full_simp_tac
								 (ocl_ss addsimps (simps)))])
	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end
	    
	fun get_mk_class_id class parent thy =
	    (* get_mk_A_id_0: "is_A_0 obj ==> get_A_0(mk_A_0 obj) = obj"  *)
	    let
		val name = "get_mk_"^(local_name_of class)^"_id"
		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_class_of class))),dummyT)$Free("obj",dummyT)) 
					  (Const("op =",dummyT)$(
					   Const((string_of_path (hc_path l0 (get_class_of class))),dummyT)
						$(Const(string_of_path (hc_path l0 (mk_class_of class)),dummyT)$Free("obj",dummyT)))
						$(Free("obj",dummyT)))

		val _ = trace 9  ("proving "^name)
		val prems_str = (string_of_path (hc_path l0 (is_class_of class)))
				::(string_of_path (hc_path l0 (get_class_of class)))
				::(string_of_path (hc_path l0 (mk_class_of class)))
				::[]
		val prems= map (fn def => get_thm thy (Name (def^"_def"))) prems_str
 
		val thm =  quick_and_dirty_prove_goalw_cterm thy [] goal_i
				       (fn p => [cut_facts_tac p 1, 
						 asm_full_simp_tac (ocl_ss addsimps prems) 1,
						 TRY(stac (get_thm thy (Name get_mk_parent)) 1),
						 TRY(asm_full_simp_tac (ocl_ss addsimps [is_parent_def]) 1),
						 ext_DEF_tac thy,
						 
						 stac (get_thm thy (Name "base_snd_drop")) 1,
						 atac 1,
						 forward_tac [get_thm (thy) (Name "DEF_baseD")]1,
				
						 asm_full_simp_tac def_X_up_ss 1,
						 REPEAT(etac exE 1),
						 asm_full_simp_tac sup_ss 1,
						 auto_tac (ocl_cla,ocl_ss)
				       ])   
	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end

	fun mk_get_class_id class parent thy =
	    (* mk_get_A_id_0: "is_A_uni_0 univ ==> mk_A_0(get_A_0 univ) = univ"  *)
	    let
		val name = "mk_get_"^(local_name_of class)^"_id"
		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_univ_class_of class))),dummyT)$Free("univ",dummyT)) 
					  (Const("op =",dummyT)$(
					   Const((string_of_path (hc_path l0 (mk_class_of class))),dummyT)
						$(Const(string_of_path (hc_path l0 (get_class_of class)),dummyT)$Free("univ",dummyT)))
						$(Free("univ",dummyT)))
		val _ = trace 9  ("proving "^name)
		val prems_str = (string_of_path (hc_path l0 (is_univ_class_of class)))
				::(string_of_path (hc_path l0 (mk_class_of class)))
				::(string_of_path (hc_path l0 (get_class_of class)))
				::("sup")::[]
		val prems= map (fn def => get_thm thy (Name (def^"_def"))) 
						prems_str
		val thm =  quick_and_dirty_prove_goalw_cterm thy [] goal_i
				       (fn p => [cut_facts_tac p 1, 
						 asm_full_simp_tac (ocl_ss addsimps prems) 1,
						 etac conjE 1,
						 ext_DEF_tac thy,
						 asm_full_simp_tac def_X_up_ss 1,
						 REPEAT(etac exE 1),	
						 asm_full_simp_tac
						     (ocl_ss) 1,
						 auto_tac (ocl_cla,ocl_ss),
						 case_tac "a" 1,
						 dtac sym 1, 
						 asm_full_simp_tac (ocl_ss
									addsimps parent_simps) 1]
				       )
	    in
		(fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end


	fun cast_class_id class parent thy =
	    (*   cast_C_id_0: "is_C_0 obj ==> (A_2_C_0(C_2_A_0(obj))) = obj" *)
	    let
		val name = "cast_"^(local_name_of class)^"_id"

		val goal_i = mkGoal_cterm (Const((string_of_path (hc_path l0 (is_class_of class))),dummyT)$Free("obj",dummyT)) 
					  (Const("op =",dummyT)$(
					   Const((string_of_path (hc_path l0 (parent2class_of class (local_name_of parent)))),dummyT)
						$(Const(string_of_path (hc_path l0 (class2get_parent class (local_name_of parent))),dummyT)
						       $Free("obj",dummyT)))
						$(Free("obj",dummyT)))


		val _ = trace 9  ("proving "^name)
		val thm =  quick_and_dirty_prove_goalw_cterm thy [] goal_i
				       (fn p => [cut_facts_tac p 1,  
						 asm_full_simp_tac 
						     (HOL_ss addsimps 
							     [o_def,
							      get_def thy (string_of_path 
									       (hc_path l0 
											(parent2class_of 
											     class 
											     (local_name_of parent)))),
							      get_def thy (string_of_path 
									       (hc_path 
										    l0 
										    (class2get_parent 
											 class 
											 (local_name_of parent))))
						     ]) 1 ,
						 
						 stac (get_thm thy (Name mk_get_parent)) 1,
					 	 asm_full_simp_tac (
						 HOL_ss addsimps [
						 get_def thy (string_of_path (hc_path l0 (is_class_of class))),
						 get_thm thy (Name ("is_"^(local_name_of parent)^"_mk_"^(local_name_of class)))
						 ] 
									) 1 ,
						 stac (get_thm thy (Name ("get_mk_"^(local_name_of class)^"_id"))) 1 ,
						 ALLGOALS(simp_tac (HOL_ss) )
  				       ])
	    in
		(fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end  
  	val thy = is_class_univ_implies_parent_univ class parent thy 
	val thy = is_class_univ_implies_is_get class parent thy 
	val thy = get_mk_class_id  class parent thy   
	val thy = is_mk_class class  parent thy
	val thy = is_parent_mk_class class  parent thy 
	val thy = mk_get_class_id class parent thy  
	val thy = cast_class_id class parent thy  		  
	val (_,thy) = set_thy_path oldpath thy
   in
	thy
    end



fun build_universe base_universe cl thy =
let
	val _ = trace 2 "    Building universe" 
	val thy_name = Context.theory_name thy
	val thy = foldl (fn (c,thy) => add_classifier_key thy c) thy cl
 	val thy = foldl (fn (c,thy) => add_classifier_type thy c) thy cl 
	val universe = foldl (fn (c,u) => insert_classifier_in_universe 
					      u thy_name c) 
			     base_universe cl
	val thy = add_type_of_universe universe thy	
 	val thy = foldl (fn (c,thy) => add_class_type 
					   thy  universe c) thy cl 
in
    (thy,universe)
end

fun add_level0 universe cl thy = 
    let													     
	val _ = trace 2 "    Adding constructors" 
	val thy = foldl (fn (c,thy) => add_constructors_0 
					   c universe cl thy) 
			thy cl   	      
	val _ = trace 2 "    Adding attributes" 
	val thy = foldl (fn (c,thy) => add_attributes_0
					   c universe cl thy) 
		thy cl   	      

	val _ = trace 2 "    Proving ..." 
	val thy = foldl (fn (c,thy) => prove_0 
					   c universe  cl  thy) 
			thy cl   	      
    in
	thy
    end
end
